# torchcell/adapters/costanzo2016_adapter.py
# [[torchcell.adapters.costanzo2016_adapter]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/adapters/costanzo2016_adapter.py
# Test file: tests/torchcell/adapters/test_costanzo2016_adapter.py

from tqdm import tqdm
import hashlib
import json
from biocypher import BioCypher
from biocypher._create import BioCypherEdge, BioCypherNode
from biocypher._logger import get_logger
import logging
from typing import Generator, Set
import torch
from torchcell.datasets.scerevisiae import (
    SmfCostanzo2016Dataset,
    DmfCostanzo2016Dataset,
)
from torchcell.datamodels import Genotype
from concurrent.futures import ProcessPoolExecutor, as_completed
from torch_geometric.loader import DataLoader
from torchcell.dataset import Dataset
from concurrent.futures import ProcessPoolExecutor
import multiprocessing as mp
from functools import partial
from typing import Callable
from functools import wraps
from abc import abstractmethod
from omegaconf import OmegaConf
from typing import Any
from torchcell.adapters.cell_adapter import CellAdapter

# logging
# Get the biocypher logger
logger = get_logger("biocypher")
logger.setLevel(logging.ERROR)


class SmfCostanzo2016Adapter(CellAdapter):
    def __init__(
        self,
        dataset: SmfCostanzo2016Dataset,
        compute_workers: int,
        io_workers: int,
        chunk_size: int = int(1e4),
        loader_batch_size: int = int(1e3),
    ):
        super().__init__(
            dataset, compute_workers, io_workers, chunk_size, loader_batch_size
        )
        self.dataset = dataset
        self.compute_workers = compute_workers
        self.io_workers = io_workers
        self.chunk_size = chunk_size
        self.loader_batch_size = loader_batch_size


class DmfCostanzo2016Adapter(CellAdapter):
    def __init__(
        self,
        dataset: DmfCostanzo2016Dataset,
        compute_workers: int,
        io_workers: int,
        chunk_size: int = int(1e4),
        loader_batch_size: int = int(1e3),
    ):
        super().__init__(
            dataset, compute_workers, io_workers, chunk_size, loader_batch_size
        )
        self.dataset = dataset
        self.compute_workers = compute_workers
        self.io_workers = io_workers
        self.chunk_size = chunk_size
        self.loader_batch_size = loader_batch_size


# class DmfCostanzo2016Adapter:
#     def __init__(
#         self,
#         dataset: DmfCostanzo2016Dataset,
#         compute_workers: int = 1,
#         chunk_size: int = 1000,
#     ):
#         self.dataset = dataset
#         self.compute_workers = compute_workers
#         self.chunk_size = chunk_size

#     def get_nodes(self) -> Generator[BioCypherNode, None, None]:
#         for node in self._get_experiment_nodes():
#             yield node
#         for node in self._get__genotype_nodes():
#             yield node
#         for node in self._get__perturbation_nodes():
#             yield node
#         for node in self._get_phenotype_nodes():
#             yield node

#         # methods = [
#         #     self._get_experiment_reference_nodes,
#         #     self._get_genome_nodes,
#         #     self._get_dataset_nodes,
#         #     self._get_environment_nodes,
#         #     self._get_media_nodes,
#         #     self._get_temperature_nodes,
#         # ]

#         # with ProcessPoolExecutor(max_workers=self.compute_workers) as executor:
#         #     futures = [executor.submit(method) for method in methods]
#         #     for future in as_completed(futures):
#         #         try:
#         #             node_generator = future.result()
#         #             for node in node_generator:
#         #                 yield node
#         #         except Exception as exc:
#         #             logger.error(
#         #                 f"Node generation method generated an exception: {exc}"
#         #             )

#     def _get_experiment_reference_nodes(self) -> list[BioCypherNode]:
#         nodes = []
#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             experiment_ref_id = hashlib.sha256(
#                 json.dumps(data.reference.model_dump()).encode("utf-8")
#             ).hexdigest()
#             node = BioCypherNode(
#                 node_id=experiment_ref_id,
#                 preferred_id=f"DmfCostanzo2016_reference_{i}",
#                 node_label="experiment reference",
#                 properties={
#                     "dataset_index": i,
#                     "serialized_data": json.dumps(data.reference.model_dump()),
#                 },
#             )
#             nodes.append(node)
#         return nodes

#     def _get_genome_nodes(self) -> list[BioCypherNode]:
#         nodes = []
#         seen_node_ids: Set[str] = set()
#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             genome_id = hashlib.sha256(
#                 json.dumps(data.reference.reference_genome.model_dump()).encode("utf-8")
#             ).hexdigest()
#             if genome_id not in seen_node_ids:
#                 seen_node_ids.add(genome_id)
#                 node = BioCypherNode(
#                     node_id=genome_id,
#                     preferred_id=f"reference_genome_{i}",
#                     node_label="genome",
#                     properties={
#                         "species": data.reference.reference_genome.species,
#                         "strain": data.reference.reference_genome.strain,
#                         "serialized_data": json.dumps(
#                             data.reference.reference_genome.model_dump()
#                         ),
#                     },
#                 )
#                 nodes.append(node)
#         return nodes

#     @staticmethod
#     def _chunk_experiment_nodes(data_chunk: dict) -> list[BioCypherNode]:
#         nodes = []
#         for i, data in enumerate(data_chunk):
#             experiment_id = hashlib.sha256(
#                 json.dumps(data["experiment"].model_dump()).encode("utf-8")
#             ).hexdigest()
#             node = BioCypherNode(
#                 node_id=experiment_id,
#                 preferred_id=f"DmfCostanzo2016_{i}",
#                 node_label="experiment",
#                 properties={
#                     "dataset_index": i,
#                     "serialized_data": json.dumps(data["experiment"].model_dump()),
#                 },
#             )
#             nodes.append(node)
#         return nodes

#     def _get_experiment_nodes(self) -> list[BioCypherNode]:
#         data_chunks = [
#             self.dataset[i : i + self.chunk_size]
#             for i in range(0, len(self.dataset), self.chunk_size)
#         ]
#         nodes = []
#         with ProcessPoolExecutor(max_workers=self.compute_workers) as executor:
#             futures = [
#                 executor.submit(self._chunk_experiment_nodes, chunk)
#                 for chunk in data_chunks
#             ]
#             for future in futures:
#                 nodes.extend(future.result())
#         return nodes

#     def _get__genotype_nodes(self) -> list[BioCypherNode]:
#         nodes = []
#         seen_node_ids = set()
#         for i, data in tqdm(enumerate(self.dataset)):
#             genotype = data["experiment"].genotype
#             genotype_id = hashlib.sha256(
#                 json.dumps(genotype.model_dump()).encode("utf-8")
#             ).hexdigest()
#             if genotype_id not in seen_node_ids:
#                 seen_node_ids.add(genotype_id)
#                 node = BioCypherNode(
#                     node_id=genotype_id,
#                     preferred_id=f"genotype_{i}",
#                     node_label="genotype",
#                     properties={
#                         "systematic_gene_names": genotype.systematic_gene_names,
#                         "perturbed_gene_names": genotype.perturbed_gene_names,
#                         "perturbation_types": genotype.perturbation_types,
#                         "serialized_data": json.dumps(
#                             data["experiment"].genotype.model_dump()
#                         ),
#                     },
#                 )
#                 nodes.append(node)
#         return nodes

#     def _get__perturbation_nodes(self) -> list[BioCypherNode]:
#         nodes = []
#         seen_node_ids = set()
#         for data in tqdm(self.dataset):
#             perturbations = data["experiment"].genotype.perturbations
#             for perturbation in perturbations:
#                 perturbation_id = hashlib.sha256(
#                     json.dumps(perturbation.model_dump()).encode("utf-8")
#                 ).hexdigest()
#                 if perturbation_id not in seen_node_ids:
#                     seen_node_ids.add(perturbation_id)
#                     node = BioCypherNode(
#                         node_id=perturbation_id,
#                         preferred_id=perturbation.perturbation_type,
#                         node_label="perturbation",
#                         properties={
#                             "systematic_gene_name": perturbation.systematic_gene_name,
#                             "perturbed_gene_name": perturbation.perturbed_gene_name,
#                             "perturbation_type": perturbation.perturbation_type,
#                             "description": perturbation.description,
#                             "strain_id": perturbation.strain_id,
#                             "serialized_data": json.dumps(perturbation.model_dump()),
#                         },
#                     )
#                     nodes.append(node)
#         return nodes

#     def _get_environment_nodes(self) -> list[BioCypherNode]:
#         # HACK - we know we can loop ref for this node type
#         nodes = []
#         seen_node_ids = set()
#         for data in tqdm(self.dataset.experiment_reference_index):
#             environment_id = hashlib.sha256(
#                 json.dumps(data.reference.reference_environment.model_dump()).encode(
#                     "utf-8"
#                 )
#             ).hexdigest()
#             if environment_id not in seen_node_ids:
#                 seen_node_ids.add(environment_id)
#                 media = json.dumps(
#                     data.reference.reference_environment.media.model_dump()
#                 )
#                 node = BioCypherNode(
#                     node_id=environment_id,
#                     preferred_id=f"environment_ref_{environment_id}",
#                     node_label="environment",
#                     properties={
#                         "temperature": data.reference.reference_environment.temperature.value,
#                         "media": media,
#                         "serialized_data": json.dumps(
#                             data.reference.reference_environment.model_dump()
#                         ),
#                     },
#                 )
#                 nodes.append(node)
#         return nodes

#     def _get_media_nodes(self) -> list[BioCypherNode]:
#         # HACK - we know we can loop ref for this node type
#         seen_node_ids = set()
#         nodes = []
#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             media_id = hashlib.sha256(
#                 json.dumps(
#                     data.reference.reference_environment.media.model_dump()
#                 ).encode("utf-8")
#             ).hexdigest()
#             if media_id not in seen_node_ids:
#                 seen_node_ids.add(media_id)
#                 name = data.reference.reference_environment.media.name
#                 state = data.reference.reference_environment.media.state
#                 node = BioCypherNode(
#                     node_id=media_id,
#                     preferred_id=f"media_{media_id}",
#                     node_label="media",
#                     properties={
#                         "name": name,
#                         "state": state,
#                         "serialized_data": json.dumps(
#                             data.reference.reference_environment.media.model_dump()
#                         ),
#                     },
#                 )
#                 nodes.append(node)
#         return nodes

#     def _get_temperature_nodes(self) -> list[BioCypherNode]:
#         # HACK - we know we can loop ref for this node type
#         seen_node_ids = set()
#         nodes = []
#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             temperature_id = hashlib.sha256(
#                 json.dumps(
#                     data.reference.reference_environment.temperature.model_dump()
#                 ).encode("utf-8")
#             ).hexdigest()
#             if temperature_id not in seen_node_ids:
#                 seen_node_ids.add(temperature_id)
#                 node = BioCypherNode(
#                     node_id=temperature_id,
#                     preferred_id=f"temperature_{temperature_id}",
#                     node_label="temperature",
#                     properties={
#                         "value": data.reference.reference_environment.temperature.value,
#                         "unit": data.reference.reference_environment.temperature.unit,
#                         "serialized_data": json.dumps(
#                             data.reference.reference_environment.temperature.model_dump()
#                         ),
#                     },
#                 )
#                 nodes.append(node)
#         return nodes

#     def _get_phenotype_nodes(self) -> list[BioCypherNode]:
#         nodes = []
#         seen_node_ids = set()
#         for i, data in tqdm(enumerate(self.dataset)):
#             phenotype_id = hashlib.sha256(
#                 json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
#             ).hexdigest()
#             if phenotype_id not in seen_node_ids:
#                 seen_node_ids.add(phenotype_id)
#                 graph_level = data["experiment"].phenotype.graph_level
#                 label = data["experiment"].phenotype.label
#                 label_error = data["experiment"].phenotype.label_error
#                 fitness = data["experiment"].phenotype.fitness
#                 fitness_std = data["experiment"].phenotype.fitness_std

#                 node = BioCypherNode(
#                     node_id=phenotype_id,
#                     preferred_id=f"phenotype_{phenotype_id}",
#                     node_label="phenotype",
#                     properties={
#                         "graph_level": graph_level,
#                         "label": label,
#                         "label_error": label_error,
#                         "fitness": fitness,
#                         "fitness_std": fitness_std,
#                         "serialized_data": json.dumps(
#                             data["experiment"].phenotype.model_dump()
#                         ),
#                     },
#                 )
#                 nodes.append(node)

#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             phenotype_id = hashlib.sha256(
#                 json.dumps(data.reference.reference_phenotype.model_dump()).encode(
#                     "utf-8"
#                 )
#             ).hexdigest()
#             if phenotype_id not in seen_node_ids:
#                 seen_node_ids.add(phenotype_id)
#                 graph_level = data.reference.reference_phenotype.graph_level
#                 label = data.reference.reference_phenotype.label
#                 label_error = data.reference.reference_phenotype.label_error
#                 fitness = data.reference.reference_phenotype.fitness
#                 fitness_std = data.reference.reference_phenotype.fitness_std

#                 node = BioCypherNode(
#                     node_id=phenotype_id,
#                     preferred_id=f"phenotype_{phenotype_id}",
#                     node_label="phenotype",
#                     properties={
#                         "graph_level": graph_level,
#                         "label": label,
#                         "label_error": label_error,
#                         "fitness": fitness,
#                         "fitness_std": fitness_std,
#                         "serialized_data": json.dumps(
#                             data.reference.reference_phenotype.model_dump()
#                         ),
#                     },
#                 )
#                 nodes.append(node)
#         return nodes

#     def _get_dataset_nodes(self) -> list[BioCypherNode]:
#         nodes = [
#             BioCypherNode(
#                 node_id="DmfCostanzo2016",
#                 preferred_id="DmfCostanzo2016",
#                 node_label="dataset",
#             )
#         ]
#         return nodes

#     def get_edges(self) -> Generator[BioCypherEdge, None, None]:
#         pass
#         # for edge in self._get_experiment_dataset_edges():
#         #     yield edge
#         # for edge in self._get_dataset_experiment_ref_edges():
#         #     yield edge
#         # for edge in self._get_experiment_ref_experiment_edges():
#         #     yield edge
#         # for edge in self._get_genotype_experiment_edges():
#         #     yield edge
#         # for edge in self._get_environment_experiment_edges():
#         #     yield edge
#         # for edge in self._get_environment_experiment_ref_edges():
#         #     yield edge
#         # for edge in self._get_phenotype_experiment_edges():
#         #     yield edge
#         # for edge in self._get_phenotype_experiment_ref_edges():
#         #     yield edge
#         # for edge in self._get_perturbation_genotype_edges():
#         #     yield edge
#         # for edge in self._get_media_environment_edges():
#         #     yield edge
#         # for edge in self._get_temperature_environment_edges():
#         #     yield edge
#         # for edge in self._get_genome_edges():
#         #     yield edge

#     # FLAG
#     @staticmethod
#     def _chunk_experiment_dataset_edges(data_chunk) -> list[BioCypherEdge]:
#         # Process a chunk of data
#         edges = []
#         for data in data_chunk:
#             experiment_id = hashlib.sha256(
#                 json.dumps(data["experiment"].model_dump()).encode("utf-8")
#             ).hexdigest()
#             edge = BioCypherEdge(
#                 source_id=experiment_id,
#                 target_id="DmfCostanzo2016",
#                 relationship_label="experiment member of",
#             )
#             edges.append(edge)
#         return edges

#     # FLAG
#     def _get_experiment_dataset_edges(self) -> list[BioCypherEdge]:
#         data_chunks = [
#             self.dataset[i : i + self.chunk_size]
#             for i in range(0, len(self.dataset), self.chunk_size)
#         ]
#         edges = []
#         with ProcessPoolExecutor(max_workers=self.compute_workers) as executor:
#             futures = [
#                 executor.submit(self._chunk_experiment_dataset_edges, chunk)
#                 for chunk in data_chunks
#             ]
#             for future in futures:
#                 edges.extend(future.result())
#         return edges

#     # FLAG
#     def _get_dataset_experiment_ref_edges(self) -> list[BioCypherEdge]:
#         edges = []
#         for data in tqdm(self.dataset.experiment_reference_index):
#             experiment_ref_id = hashlib.sha256(
#                 json.dumps(data.reference.model_dump()).encode("utf-8")
#             ).hexdigest()
#             edge = BioCypherEdge(
#                 source_id=experiment_ref_id,
#                 target_id="DmfCostanzo2016",
#                 relationship_label="experiment reference member of",
#             )
#             edges.append(edge)
#         return edges

#     # FLAG
#     def _get_experiment_ref_experiment_edges(self) -> list[BioCypherEdge]:
#         edges = []
#         for data in tqdm(self.dataset.experiment_reference_index):
#             dataset_subset = self.dataset[torch.tensor(data.index)]
#             experiment_ref_id = hashlib.sha256(
#                 json.dumps(data.reference.model_dump()).encode("utf-8")
#             ).hexdigest()
#             for i, data in enumerate(dataset_subset):
#                 experiment_id = hashlib.sha256(
#                     json.dumps(data["experiment"].model_dump()).encode("utf-8")
#                 ).hexdigest()
#                 edge = BioCypherEdge(
#                     source_id=experiment_ref_id,
#                     target_id=experiment_id,
#                     relationship_label="experiment reference of",
#                 )
#                 edges.append(edge)
#         return edges

#     # FLAG
#     @staticmethod
#     def _chunk_genotype_experiment_edges(data_chunk: dict) -> list[BioCypherEdge]:
#         edges = []
#         for data in data_chunk:
#             experiment_id = hashlib.sha256(
#                 json.dumps(data["experiment"].model_dump()).encode("utf-8")
#             ).hexdigest()
#             genotype = data["experiment"].genotype
#             genotype_id = hashlib.sha256(
#                 json.dumps(genotype.model_dump()).encode("utf-8")
#             ).hexdigest()
#             edge = BioCypherEdge(
#                 source_id=genotype_id,
#                 target_id=experiment_id,
#                 relationship_label="genotype member of",
#             )
#             edges.append(edge)
#         return edges

#     # FLAG
#     def _get_genotype_experiment_edges(self) -> list[BioCypherEdge]:
#         data_chunks = [
#             self.dataset[i : i + self.chunk_size]
#             for i in range(0, len(self.dataset), self.chunk_size)
#         ]
#         edges = []
#         with ProcessPoolExecutor(max_workers=self.compute_workers) as executor:
#             futures = [
#                 executor.submit(self._chunk_genotype_experiment_edges, chunk)
#                 for chunk in data_chunks
#             ]
#             for future in futures:
#                 edges.extend(future.result())
#         return edges

#     # FLAG
#     def _get_perturbation_genotype_edges(self) -> list[BioCypherEdge]:
#         edges = []
#         seen_edges = set()
#         for data in tqdm(self.dataset):
#             genotype = data["experiment"].genotype
#             for perturbation in genotype.perturbations:
#                 genotype_id = hashlib.sha256(
#                     json.dumps(genotype.model_dump()).encode("utf-8")
#                 ).hexdigest()
#                 perturbation_id = hashlib.sha256(
#                     json.dumps(perturbation.model_dump()).encode("utf-8")
#                 ).hexdigest()
#                 edge_tuple = (perturbation_id, genotype_id, "perturbation member of")
#                 if edge_tuple not in seen_edges:
#                     seen_edges.add(edge_tuple)
#                     edge = BioCypherEdge(
#                         source_id=perturbation_id,
#                         target_id=genotype_id,
#                         relationship_label="perturbation member of",
#                     )
#                     edges.append(edge)

#         return edges

#     # FLAG
#     @staticmethod
#     def _chunk_environment_experiment_edges(data_chunk) -> list[BioCypherEdge]:
#         edges = []
#         seen_environment_experiment_pairs = set()
#         for data in data_chunk:
#             experiment_id = hashlib.sha256(
#                 json.dumps(data["experiment"].model_dump()).encode("utf-8")
#             ).hexdigest()
#             environment_id = hashlib.sha256(
#                 json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
#             ).hexdigest()
#             env_experiment_pair = (environment_id, experiment_id)
#             if env_experiment_pair not in seen_environment_experiment_pairs:
#                 seen_environment_experiment_pairs.add(env_experiment_pair)
#                 edge = BioCypherEdge(
#                     source_id=environment_id,
#                     target_id=experiment_id,
#                     relationship_label="environment member of",
#                 )
#                 edges.append(edge)
#         return edges

#     # FLAG
#     def _get_environment_experiment_edges(self) -> list[BioCypherEdge]:
#         data_chunks = [
#             self.dataset[i : i + self.chunk_size]
#             for i in range(0, len(self.dataset), self.chunk_size)
#         ]
#         edges = []
#         with ProcessPoolExecutor(max_workers=self.compute_workers) as executor:
#             futures = [
#                 executor.submit(self._chunk_environment_experiment_edges, chunk)
#                 for chunk in data_chunks
#             ]
#             for future in futures:
#                 edges.extend(future.result())
#         return edges

#     # FLAG
#     def _get_environment_experiment_ref_edges(self) -> list[BioCypherEdge]:
#         edges = []
#         seen_environment_experiment_ref_pairs: Set[tuple] = set()
#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             experiment_ref_id = hashlib.sha256(
#                 json.dumps(data.reference.model_dump()).encode("utf-8")
#             ).hexdigest()
#             environment_id = hashlib.sha256(
#                 json.dumps(data.reference.reference_environment.model_dump()).encode(
#                     "utf-8"
#                 )
#             ).hexdigest()
#             env_experiment_ref_pair = (environment_id, experiment_ref_id)
#             if env_experiment_ref_pair not in seen_environment_experiment_ref_pairs:
#                 seen_environment_experiment_ref_pairs.add(env_experiment_ref_pair)

#                 edge = BioCypherEdge(
#                     source_id=environment_id,
#                     target_id=experiment_ref_id,
#                     relationship_label="environment member of",
#                 )
#                 edges.append(edge)
#         return edges

#     # FLAG
#     @staticmethod
#     def _chunk_phenotype_experiment_edges(data_chunk) -> list[BioCypherEdge]:
#         edges = []
#         seen_phenotype_experiment_pairs = set()
#         for data in data_chunk:
#             experiment_id = hashlib.sha256(
#                 json.dumps(data["experiment"].model_dump()).encode("utf-8")
#             ).hexdigest()
#             phenotype_id = hashlib.sha256(
#                 json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
#             ).hexdigest()
#             phenotype_experiment_pair = (phenotype_id, experiment_id)
#             if phenotype_experiment_pair not in seen_phenotype_experiment_pairs:
#                 seen_phenotype_experiment_pairs.add(phenotype_experiment_pair)
#                 edge = BioCypherEdge(
#                     source_id=phenotype_id,
#                     target_id=experiment_id,
#                     relationship_label="phenotype member of",
#                 )
#                 edges.append(edge)
#         return edges

#     # FLAG
#     def _get_phenotype_experiment_edges(self) -> list[BioCypherEdge]:
#         data_chunks = [
#             self.dataset[i : i + self.chunk_size]
#             for i in range(0, len(self.dataset), self.chunk_size)
#         ]
#         edges = []
#         with ProcessPoolExecutor(max_workers=self.compute_workers) as executor:
#             futures = [
#                 executor.submit(self._chunk_phenotype_experiment_edges, chunk)
#                 for chunk in data_chunks
#             ]
#             for future in futures:
#                 edges.extend(future.result())
#         return edges

#     # FLAG
#     def _get_phenotype_experiment_ref_edges(self) -> list[BioCypherEdge]:
#         edges = []
#         seen_phenotype_experiment_ref_pairs: Set[tuple] = set()
#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             experiment_ref_id = hashlib.sha256(
#                 json.dumps(data.reference.model_dump()).encode("utf-8")
#             ).hexdigest()
#             phenotype_id = hashlib.sha256(
#                 json.dumps(data.reference.reference_phenotype.model_dump()).encode(
#                     "utf-8"
#                 )
#             ).hexdigest()
#             phenotype_experiment_ref_pair = (phenotype_id, experiment_ref_id)
#             if phenotype_experiment_ref_pair not in seen_phenotype_experiment_ref_pairs:
#                 seen_phenotype_experiment_ref_pairs.add(phenotype_experiment_ref_pair)

#                 edge = BioCypherEdge(
#                     source_id=phenotype_id,
#                     target_id=experiment_ref_id,
#                     relationship_label="phenotype member of",
#                 )
#                 edges.append(edge)
#         return edges

#     def _get_media_environment_edges(self) -> list[BioCypherEdge]:
#         # HACK Optimized by using reference
#         # We know reference contains all media and envs
#         edges = []
#         seen_media_environment_pairs: Set[tuple] = set()
#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             environment_id = hashlib.sha256(
#                 json.dumps(data.reference.reference_environment.model_dump()).encode(
#                     "utf-8"
#                 )
#             ).hexdigest()
#             media_id = hashlib.sha256(
#                 json.dumps(
#                     data.reference.reference_environment.media.model_dump()
#                 ).encode("utf-8")
#             ).hexdigest()
#             media_environment_pair = (media_id, environment_id)
#             if media_environment_pair not in seen_media_environment_pairs:
#                 seen_media_environment_pairs.add(media_environment_pair)
#                 edge = BioCypherEdge(
#                     source_id=media_id,
#                     target_id=environment_id,
#                     relationship_label="media member of",
#                 )
#                 edges.append(edge)
#         return edges

#     def _get_temperature_environment_edges(self) -> list[BioCypherEdge]:
#         # HACK Optimized by using reference
#         # We know reference contain all envs and temps
#         edges = []
#         seen_temperature_environment_pairs: Set[tuple] = set()
#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             environment_id = hashlib.sha256(
#                 json.dumps(data.reference.reference_environment.model_dump()).encode(
#                     "utf-8"
#                 )
#             ).hexdigest()
#             temperature_id = hashlib.sha256(
#                 json.dumps(
#                     data.reference.reference_environment.temperature.model_dump()
#                 ).encode("utf-8")
#             ).hexdigest()
#             temperature_environment_pair = (temperature_id, environment_id)
#             if temperature_environment_pair not in seen_temperature_environment_pairs:
#                 seen_temperature_environment_pairs.add(temperature_environment_pair)
#                 edge = BioCypherEdge(
#                     source_id=temperature_id,
#                     target_id=environment_id,
#                     relationship_label="temperature member of",
#                 )
#                 edges.append(edge)
#         return edges

#     def _get_genome_edges(self) -> list[BioCypherEdge]:
#         edges = []
#         seen_genome_experiment_ref_pairs: Set[tuple] = set()
#         for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
#             experiment_ref_id = hashlib.sha256(
#                 json.dumps(data.reference.model_dump()).encode("utf-8")
#             ).hexdigest()
#             genome_id = hashlib.sha256(
#                 json.dumps(data.reference.reference_genome.model_dump()).encode("utf-8")
#             ).hexdigest()
#             genome_experiment_ref_pair = (genome_id, experiment_ref_id)
#             if genome_experiment_ref_pair not in seen_genome_experiment_ref_pairs:
#                 seen_genome_experiment_ref_pairs.add(genome_experiment_ref_pair)

#                 edge = BioCypherEdge(
#                     source_id=genome_id,
#                     target_id=experiment_ref_id,
#                     relationship_label="genome member of",
#                 )
#                 edges.append(edge)
#         return edges


if __name__ == "__main__":
    import os.path as osp
    from dotenv import load_dotenv
    import math
    from datetime import datetime
    import os

    ##
    load_dotenv()
    time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    DATA_ROOT = os.getenv("DATA_ROOT")
    BIOCYPHER_CONFIG_PATH = os.getenv("BIOCYPHER_CONFIG_PATH")
    SCHEMA_CONFIG_PATH = os.getenv("SCHEMA_CONFIG_PATH")

    ## SMF
    # bc = BioCypher(
    #     output_directory=osp.join(DATA_ROOT, "database/biocypher-out", time),
    #     biocypher_config_path=BIOCYPHER_CONFIG_PATH,
    #     schema_config_path=SCHEMA_CONFIG_PATH,
    # )
    # dataset = SmfCostanzo2016Dataset(
    #     osp.join(DATA_ROOT, "data/torchcell/smf_costanzo2016")
    # )
    # num_workers = mp.cpu_count()
    # io_workers = math.ceil(0.2 * num_workers)
    # compute_workers = num_workers - io_workers
    # adapter = SmfCostanzo2016Adapter(
    #     dataset=dataset,
    #     compute_workers=6,
    #     io_workers=4,
    #     chunk_size=int(1e4),
    #     loader_batch_size=int(1e4),
    # )
    # bc.write_nodes(adapter.get_nodes())
    # bc.write_edges(adapter.get_edges())
    # bc.write_import_call()
    # bc.write_schema_info(as_node=True)
    # bc.summary()

    ## DMF
    bc = BioCypher(
        output_directory=osp.join(DATA_ROOT, "database/biocypher-out", time),
        biocypher_config_path=BIOCYPHER_CONFIG_PATH,
        schema_config_path=SCHEMA_CONFIG_PATH,
    )
    # dataset = DmfCostanzo2016Dataset(
    #     root=osp.join(DATA_ROOT, "data/torchcell/dmf_costanzo2016")
    # )
    dataset = DmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/torchcell/dmf_costanzo2016_subset_n_1e3"),
        subset_n=int(1e3),
    )
    adapter = DmfCostanzo2016Adapter(
        dataset=dataset,
        compute_workers=1,
        io_workers=1,
        chunk_size=int(2 * 1e2),
        loader_batch_size=int(1e2),
    )
    bc.write_nodes(adapter.get_nodes())
    # bc.write_edges(adapter.get_edges())
    bc.write_import_call()
    bc.write_schema_info(as_node=True)
    bc.summary()



import os
import argparse
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor

def delete_file(file_path):
    if os.path.exists(file_path):  # Check if file exists
        os.remove(file_path)
    else:
        print(f"File {file_path} does not exist.")

def delete_files(files):
    with ThreadPoolExecutor() as executor:
        list(executor.map(delete_file, files), total=len(files), desc="Deleting files", unit="file")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Delete a subset of files.")
    parser.add_argument("path", help="Path to the directory containing files to delete.")
    parser.add_argument("--task-id", type=int, help="ID of this task (starting from 1).")
    parser.add_argument("--total-tasks", type=int, help="Total number of tasks.")
    args = parser.parse_args()

    folder_path = args.path
    task_id = args.task_id
    total_tasks = args.total_tasks

    if not os.path.exists(folder_path):
        print(f"The directory {folder_path} does not exist.")
        exit(1)

    all_files = []
    for root, _, files in os.walk(folder_path):
        all_files.extend([os.path.join(root, f) for f in files])

    # Divide files among tasks
    my_files = [f for i, f in enumerate(all_files) if i % total_tasks == (task_id - 1)]

    delete_files(my_files)



# torchcell/datasets/base_cell.py
# [[torchcell.datasets.base_cell]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/base_cell.py
# Test file: tests/torchcell/datasets/test_base_cell.py

import copy
import json
import logging
import os
import os.path as osp
import pickle
import re
import shutil
import threading
import zipfile
from abc import ABC, abstractmethod
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from os import environ
from typing import List, Optional, Tuple, Union

import lmdb
import networkx as nx
import numpy as np
import pandas as pd
import torch
from attrs import define
from pydantic import BaseModel, Extra, Field, ValidationError, field_validator
from sklearn import experimental
from torch_geometric.data import Batch, Data, InMemoryDataset, download_url, extract_zip
from torch_geometric.data.separate import separate
from torch_geometric.loader import DataLoader
from torch_geometric.utils import (
    add_self_loops,
    coalesce,
    from_networkx,
    k_hop_subgraph,
    subgraph,
)
from tqdm import tqdm

from torchcell.data import Dataset
from torchcell.datamodels import ModelStrictArbitrary
from torchcell.datasets.codon_frequency import CodonFrequencyDataset
from torchcell.datasets.embedding import BaseEmbeddingDataset
from torchcell.datasets.fungal_up_down_transformer import FungalUpDownTransformerDataset
from torchcell.datasets.nucleotide_transformer import NucleotideTransformerDataset
from torchcell.datasets.scerevisiae import (
    DmfCostanzo2016Dataset,
    SmfCostanzo2016Dataset,
)
from torchcell.models.llm import NucleotideModel
from torchcell.models.nucleotide_transformer import NucleotideTransformer
from torchcell.prof import prof, prof_input
from torchcell.sequence import GeneSet, Genome
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome
from attrs import define, field
from neo4j import GraphDatabase

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


@define
class BaseQuery:
    query: str
    uri: str = field(default="neo4j://localhost:7687")
    username: str = field(default="neo4j")
    password: str = field(default="neo4j")

    def __attrs_post_init__(self):
        self.driver = GraphDatabase.driver(
            self.uri, auth=(self.username, self.password)
        )

    def get_data(self):
        with self.driver.session() as session:
            result = session.run(self.query)
            for record in result:
                yield record
        self.driver.close()


class BaseCellDataset(Dataset):
    pass


class Cell(Dataset):
    def __init__(
        self,
        root: str = "data/torchcell/dmf_costanzo2016",
        subset_n: int = None,
        preprocess: dict | None = None,
        skip_process_file_exist: bool = False,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        self.subset_n = subset_n
        self._skip_process_file_exist = skip_process_file_exist
        # TODO consider moving to a well defined Dataset class
        self.preprocess = preprocess
        # TODO consider moving to Dataset
        self.preprocess_dir = osp.join(root, "preprocess")
        self._length = None
        self._gene_set = None
        self._df = None
        # Check for existing preprocess config
        existing_config = self.load_preprocess_config()
        if existing_config is not None:
            if existing_config != self.preprocess:
                raise ValueError(
                    "New preprocess does not match existing config."
                    "Delete the processed and process dir for a new Dataset."
                    "Or define a new root."
                )
        self.env = None
        self._experiment_reference_index = None
        super().__init__(root, transform, pre_transform)
        # This was here before - not sure if it has something to do with gpu
        # self.env = None

    @property
    def skip_process_file_exist(self):
        return self._skip_process_file_exist

    @property
    def raw_file_names(self) -> list[str]:
        return "dummy.txt"

    @property
    def processed_file_names(self) -> list[str]:
        return "data.lmdb"

    def download(self):
        # TODO Run query to download data
        pass

    def _init_db(self):
        """Initialize the LMDB environment."""
        self.env = lmdb.open(
            osp.join(self.processed_dir, "data.lmdb"),
            readonly=True,
            lock=False,
            readahead=False,
            meminit=False,
        )

    def close_lmdb(self):
        if self.env is not None:
            self.env.close()
            self.env = None

    @property
    def df(self):
        if osp.exists(osp.join(self.preprocess_dir, "data.csv")):
            self._df = pd.read_csv(osp.join(self.preprocess_dir, "data.csv"))
        return self._df

    def process(self):
        os.makedirs(self.preprocess_dir, exist_ok=True)
        self._length = None
        # Initialize an empty DataFrame to hold all raw data
        df = pd.DataFrame()

        # Read and concatenate all raw files
        print("Reading and Concatenating Raw Files...")
        for file_name in tqdm(self.raw_file_names):
            file_path = os.path.join(self.raw_dir, file_name)

            # Reading data using Pandas; limit rows for demonstration
            df_temp = pd.read_csv(file_path, sep="\t")

            # Concatenating data frames
            df = pd.concat([df, df_temp], ignore_index=True)
        # Functions for data filtering... duplicates selection,
        df = self.preprocess_raw(df, self.preprocess)
        self.save_preprocess_config(self.preprocess)

        # Subset
        if self.subset_n is not None:
            df = df.sample(n=self.subset_n, random_state=42).reset_index(drop=True)

        # Save preprocssed df - mainly for quick stats
        df.to_csv(osp.join(self.preprocess_dir, "data.csv"), index=False)

        print("Processing DMF Files...")

        # Initialize LMDB environment
        env = lmdb.open(
            osp.join(self.processed_dir, "data.lmdb"),
            map_size=int(1e12),  # Adjust map_size as needed
        )

        with env.begin(write=True) as txn:
            for index, row in tqdm(df.iterrows(), total=df.shape[0]):
                experiment, reference = self.create_experiment(row)

                # Serialize the Pydantic objects
                serialized_data = pickle.dumps(
                    {"experiment": experiment, "reference": reference}
                )
                txn.put(f"{index}".encode(), serialized_data)

        env.close()
        self.gene_set = self.compute_gene_set()

    @staticmethod
    def create_experiment(row):
        # return experiment, reference
        pass

    def preprocess_raw(self, df: pd.DataFrame, preprocess: dict | None = None):
        print("Preprocess on raw data...")
        # return df
        pass

    # New method to save preprocess configuration to a JSON file
    def save_preprocess_config(self, preprocess):
        if not osp.exists(self.preprocess_dir):
            os.makedirs(self.preprocess_dir)
        with open(osp.join(self.preprocess_dir, "preprocess_config.json"), "w") as f:
            json.dump(preprocess, f)

    def load_preprocess_config(self):
        config_path = osp.join(self.preprocess_dir, "preprocess_config.json")

        if osp.exists(config_path):
            with open(config_path) as f:
                config = json.load(f)
            return config
        else:
            return None

    def len(self) -> int:
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            length = txn.stat()["entries"]

        # Must be closed for dataloader num_workers > 0
        self.close_lmdb()

        return length

    def get(self, idx):
        if self.env is None:
            self._init_db()

        # Handling boolean index tensors or numpy arrays
        if isinstance(idx, (list, np.ndarray, torch.Tensor)):
            if isinstance(idx, list):
                idx = np.array(idx)
            if isinstance(idx, np.ndarray) and idx.dtype == np.bool:
                idx = np.where(idx)[0]
            elif isinstance(idx, torch.Tensor) and idx.dtype == torch.bool:
                idx = idx.nonzero(as_tuple=False).squeeze(1)

        if isinstance(idx, (np.ndarray, list, torch.Tensor)):
            # If idx is a list/array/tensor of indices, return a list of data objects
            return [self.get_single_item(i.item()) for i in idx]
        else:
            # Single item retrieval
            return self.get_single_item(idx)

    def get_single_item(self, idx):
        with self.env.begin() as txn:
            serialized_data = txn.get(f"{idx}".encode())
            if serialized_data is None:
                return None

            deserialized_data = pickle.loads(serialized_data)
            return deserialized_data

    @staticmethod
    def extract_systematic_gene_names(genotypes):
        gene_names = []
        for genotype in genotypes:
            if hasattr(genotype, "perturbation") and hasattr(
                genotype.perturbation, "systematic_gene_name"
            ):
                gene_name = genotype.perturbation.systematic_gene_name
                gene_names.append(gene_name)
        return gene_names

    def compute_gene_set(self):
        gene_set = GeneSet()
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            cursor = txn.cursor()
            print("Computing gene set...")
            for key, value in tqdm(cursor):
                deserialized_data = pickle.loads(value)
                experiment = deserialized_data["experiment"]

                extracted_gene_names = self.extract_systematic_gene_names(
                    experiment.genotype
                )
                for gene_name in extracted_gene_names:
                    gene_set.add(gene_name)

        self.close_lmdb()
        return gene_set

    # Reading from JSON and setting it to self._gene_set
    @property
    def gene_set(self):
        if osp.exists(osp.join(self.preprocess_dir, "gene_set.json")):
            with open(osp.join(self.preprocess_dir, "gene_set.json")) as f:
                self._gene_set = GeneSet(json.load(f))
        elif self._gene_set is None:
            raise ValueError(
                "gene_set not written during process. "
                "Please call compute_gene_set in process."
            )
        return self._gene_set

    @gene_set.setter
    def gene_set(self, value):
        if not value:
            raise ValueError("Cannot set an empty or None value for gene_set")
        with open(osp.join(self.preprocess_dir, "gene_set.json"), "w") as f:
            json.dump(list(sorted(value)), f, indent=0)
        self._gene_set = value

    @property
    def experiment_reference_index(self):
        # TODO implement
        self._experiment_reference_index = None
        return self._experiment_reference_index

    def __repr__(self):
        return f"{self.__class__.__name__}({len(self)})"


if __name__ == "__main__":
    pass



# torchcell/multidigraph/metabolism.py
# [[torchcell.multidigraph.metabolism]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/multidigraph/metabolism.py
# Test file: torchcell/multidigraph/test_metabolism.py

import networkx as nx
import pandas as pd

# 1. Read the Excel file
file_path = "data/sgd/genome/yeast-GEM.xlsx"
data = pd.read_excel(file_path)

# 2. Parse the dataframe
G = nx.DiGraph()  # Initialize a directed graph

for index, row in data.iterrows():
    equation = row["EQUATION"]
    if "<=>" in equation:  # bidirectional
        reactants, products = equation.split("<=>")
    elif "=>" in equation:  # one-directional
        reactants, products = equation.split("=>")
    else:
        continue

    # Attributes for the reaction node
    attributes = {
        "EC": row["EC-NUMBER"],
        "Gene Association": row["GENE ASSOCIATION"],
        "MIRIAM": row["MIRIAM"],
        "Subsystem": row["SUBSYSTEM"],
    }

    reaction_node = row["ID"]  # Node name based on the "ID" column
    G.add_node(reaction_node, **attributes)  # Add reaction node with attributes

    # Connect reactants to the reaction node
    for reactant in reactants.split("+"):
        G.add_edge(
            reactant.strip(), reaction_node, stoichiometric_coefficient=1
        )  # Assuming a default stoichiometric coefficient of 1

    # Connect the reaction node to products
    for product in products.split("+"):
        G.add_edge(
            reaction_node, product.strip(), stoichiometric_coefficient=1
        )  # Assuming a default stoichiometric coefficient of 1

print()
# If you want to visualize the graph
# import matplotlib.pyplot as plt
# nx.draw(G, with_labels=True)
# plt.show()



from .pydant import ModelStrict, ModelStrictArbitrary

from .schema import (
    BaseEnvironment,
    Genotype,
    BasePhenotype,
    BaseExperiment,
    GenePerturbation,
    Media,
    ModelStrict,
    ReferenceGenome,
    Temperature,
    DeletionPerturbation,
    FitnessPhenotype,
    FitnessExperiment,
    DampPerturbation,
    TsAllelePerturbation,
    FitnessExperimentReference,
    ExperimentReference,
    KanMxDeletionPerturbation,
    NatMxDeletionPerturbation,
    SgaKanMxDeletionPerturbation,
    SgaNatMxDeletionPerturbation,
    SgaTsAllelePerturbation,
    SgaDampPerturbation,
    SuppressorAllelePerturbation,
    SgaSuppressorAllelePerturbation,
    AllelePerturbation,
    SgaAllelePerturbation,
)

core_models = ["ModelStrict", "ModelStrictArbitrary"]
ontology_models = [
    "BaseEnvironment",
    "Genotype",
    "BasePhenotype",
    "BaseExperiment",
    "GenePerturbation",
    "Media",
    "ReferenceGenome",
    "Temperature",
    "DeletionPerturbation",
    "FitnessPhenotype",
    "FitnessExperiment",
    "DampPerturbation",
    "TsAllelePerturbation",
    "FitnessExperimentReference",
    "ExperimentReference",
    "KanMxDeletionPerturbation",
    "NatMxDeletionPerturbation",
    "SgaKanMxDeletionPerturbation",
    "SgaNatMxDeletionPerturbation",
    "SgaTsAllelePerturbation",
    "SgaDampPerturbation",
    "SuppressorAllelePerturbation",
    "SgaSuppressorAllelePerturbation",
    "AllelePerturbation",
    "SgaAllelePerturbation",
]

__all__ = core_models + ontology_models





# torchcell/sequence/data.py
# [[torchcell.sequence.data]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/sequence/data.py
# Test file: /torchcell/sequence/test_data.py

import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from itertools import product
from turtle import st
from typing import Set

import gffutils
import matplotlib.pyplot as plt
import pandas as pd
from attrs import define, field
from Bio import SeqIO
from Bio.Seq import Seq
from gffutils import Feature, FeatureDB
from gffutils.biopython_integration import to_seqfeature
from matplotlib import pyplot as plt
from pydantic import BaseModel, ConfigDict, Field, ValidationError, field_validator, root_validator
from sortedcontainers import SortedDict, SortedSet
from sympy import sequence

from torchcell.datamodels import ModelStrict, ModelStrictArbitrary

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class DnaSelectionResult(ModelStrict):
    id: str
    chromosome: int
    strand: str
    start: int
    end: int
    seq: str

    def __len__(self) -> int:
        return len(self.seq)

    def __ge__(self, other: str) -> bool:
        if isinstance(other, DnaSelectionResult):
            return len(self.seq) >= len(other.seq)
        return NotImplemented

    def __le__(self, other: str) -> bool:
        if isinstance(other, DnaSelectionResult):
            return len(self.seq) <= len(other.seq)
        return NotImplemented

    @root_validator(pre=True)
    def end_leq_start(cls, values):
        start, end = values.get("start"), values.get("end")
        if start > end:
            raise ValueError("Start must be less than end")
        return values

    @field_validator("strand")
    def check_strand(cls, v):
        if v not in ["+", "-"]:
            raise ValueError("Strand must be either '+' or '-'")
        return v

    @field_validator("chromosome", "start", "end")
    def check_positive(cls, v):
        if v < 0:
            raise ValueError(f"{v} must be positive")
        return v

    # TODO consider adding chromosome length
    @field_validator("seq")
    def check_seq_len(cls, v):
        sequence_length = len(v)
        if sequence_length < 0:
            raise ValueError(f"Sequence length ({sequence_length}) not geq 0")
        return v


class DnaWindowResult(DnaSelectionResult):
    start_window: int
    end_window: int

    def __repr__(self) -> str:
        # Use f-string to create a formatted string
        return f"DnaWindowResult(id={self.id!r}, chromosome={self.chromosome!r}, strand={self.strand!r}, start_window={self.start_window!r}, end_window={self.end_window!r}, seq={self.seq!r})"

    @field_validator("start_window", "end_window")
    def check_window(cls, v):
        if v < 0:
            raise ValueError(f"{v} must be positive")
        return v


class GeneSet(SortedSet):
    def __init__(self, iterable=None, key=None):
        super().__init__(iterable, key)
        for item in self:
            if not isinstance(item, str):
                raise ValueError(
                    f"All items in gene_set must be str, got {type(item).__name__}"
                )

    def __repr__(self):
        n = len(self)
        limited_items = (self)[:3]
        if len(self) > 3:
            return f"GeneSet(size={n}, items={limited_items}...)"
        elif len(self) < 3:
            return f"GeneSet(size={n}, items={limited_items})"


###########
# Abstract Base Class for structure


##########
# Class holding gene
class Gene(ABC):
    # model_config = ConfigDict(frozen=True, extra="forbid")
    seq: str

    @abstractmethod
    def window(self, window_size: int, is_max_size: bool = True) -> DnaWindowResult:
        pass

    @abstractmethod
    def window_five_prime(
        self, window_size: int, allow_undersize: bool = False
    ) -> DnaWindowResult:
        pass

    @abstractmethod
    def window_three_prime(
        self, window_size: int, allow_undersize: bool = False
    ) -> DnaWindowResult:
        pass

    def __len__(self) -> int:
        return len(self.seq)

    # name: str
    # seq: str
    # chromosome: int
    # start: int
    # end: int
    # strand: str
    # five_utr: str
    # three_utr: str


class Genome(ABC):
    # Used elsewhere [[torchcell/sgd/validation/valid_models.py]]
    # CHECK IF THIS IS NEEDED.. I think this is a pydantic thing
    # model_config = ConfigDict(frozen=True, extra="forbid")
    # TODO not sure if we need to specify all vars in the __init__
    # TODO do we need to set data_root like this?
    def __init__(self, data_root: str = None):
        self.data_root: str = data_root
        self.db: FeatureDB = None
        self.fasta_sequences: dict = None
        self.chr_to_nc: dict = None
        self.nc_to_chr: dict = None
        self.chr_to_len: dict = None
        self._gene_set: set[str] = None
        self._fasta_path: str = None
        self._gff_path: str = None

    @property
    def gene_set(self) -> GeneSet:
        if self._gene_set is None:
            self._gene_set = self.compute_gene_set()
        return self._gene_set

    @gene_set.setter
    def gene_set(self, value: set[str]):
        self._gene_set = value

    @abstractmethod
    def compute_gene_set(self) -> set[str]:
        pass  # Abstract methods don't have a body

    @abstractmethod
    def get_seq(
        self, chr: int | str, start: int, end: int, strand: str
    ) -> DnaSelectionResult:
        pass

    @property
    @abstractmethod
    def gene_attribute_table(self) -> pd.DataFrame:
        pass

    @property
    @abstractmethod
    def feature_types(self) -> list[str]:
        pass

    @abstractmethod
    def __getitem__(self, item: str) -> Gene | None:
        pass

    # Not sure if it makes more sense to have the number of genes be the length or the sum bp over all chromosomes.
    def __len__(self) -> int:
        return len(self.gene_set)


############
# Helper functions
def mismatch_positions(seq1: str, seq2: str) -> list[int]:
    """
    Computes the positions at which two sequences differ.

    This function takes two sequences, seq1 and seq2, represented as strings
    and returns a list of positions at which the two sequences have different
    characters. The sequences must be of the same length, else a ValueError is raised.

    Args:
        seq1 (str): The first sequence to compare.
        seq2 (str): The second sequence to compare.

    Returns:
        list[str]: A list containing the positions at
        which the two sequences differ.
        An empty list is returned if the sequences are identical.

    Raises:
        ValueError: If the lengths of seq1 and seq2 are not equal.

    Example:
        >>> mismatch_positions("ATGC", "ATCC")
        [2]
    """

    if len(seq1) != len(seq2):
        raise ValueError("Sequences must be the same length")
    mismatches = [i for i, (n1, n2) in enumerate(zip(seq1, seq2)) if n1 != n2]
    return mismatches


def get_chr_from_description(description: str) -> int:
    """
    Extracts the chromosome number from a given description string.

    Processes a description string containing either a chromosome
    in Roman numeral (e.g., "[chromosome=IX]").
    Or a location (e.g., "[location=mitochondrion]").
    If a chromosome is found, it converts the Roman numeral to an integer.
    If the location is a mitochondrion, it returns 0.
    Handles descriptions containing chromosome information in a predefined format.

    Args:
        description (str): Description string containing chromosome number

    Returns:
        int: Chromosome number.  Returns 0 if the location is mitochondrion.

    Raises:
        ValueError: If the Roman numeral conversion fails due to invalid format.

    Example:
        >>> get_chr_from_description("[chromosome=IX] some other info")
        9
        >>> get_chr_from_description("[location=mitochondrion] some other info")
        0
    """
    # CHECK - format might be specific yeast s288c genome
    desc_split = description.split()
    for part in desc_split:
        if part.startswith("[chromosome="):
            roman_num = part[
                len("[chromosome=") : -1
            ]  # Exclude the last char, which is "]"
            return roman_to_int(roman_num)
        elif part.startswith("[location="):
            # we assign mitochondiral DNA to chromosome 0 as convenience
            if part[len("[location=") : -1] == "mitochondrion":
                return 0


def roman_to_int(s: str) -> int:
    """
    Converts a Roman numeral string to an integer.

    This function interprets the given string `s` as a Roman numeral and
    returns its value as an integer. It handles the standard Roman numeral
    symbols (I, V, X, L, C, D, M) and uses the subtractive notation rule,
    where placing a smaller numeral to the left of a larger numeral
    represents subtraction (e.g., IV for 4).

    Args:
        s (str): The Roman numeral string to convert, consisting of the characters
                 I, V, X, L, C, D, M. It is assumed to be a valid Roman numeral.

    Returns:
        int: The integer value of the Roman numeral.

    Raises:
        KeyError: If input string contains characters without valid Roman numeral.

    Example:
        >>> roman_to_int("IV")
        4
        >>> roman_to_int("IX")
        9
        >>> roman_to_int("XIII")
        13

    """
    roman_to_int_mapping = {
        "I": 1,
        "V": 5,
        "X": 10,
        "L": 50,
        "C": 100,
        "D": 500,
        "M": 1000,
    }
    result = 0
    for i in range(len(s)):
        if i > 0 and roman_to_int_mapping[s[i]] > roman_to_int_mapping[s[i - 1]]:
            result += roman_to_int_mapping[s[i]] - 2 * roman_to_int_mapping[s[i - 1]]
        else:
            result += roman_to_int_mapping[s[i]]
    return result


# Selection Window functions


def calculate_window_undersized(
    start: int, end: int, strand: str, window_size: int
) -> tuple[int, int]:
    """
    Calculate the start and end points of a genomic window, respecting the given strand.

    For "+" strand, the window is created from the `start` point, and for "-" strand,
    the window is created from the `end` point. This method ensures that the resulting
    window is of the specified `window_size`, handling undersized windows in the
    process. This method is particularly useful when selecting from the start of a gene.

    Args:
        start (int): The start point of the gene on the genome.
        end (int): The end point of the gene on the genome.
        strand (str): The strand of the gene, either "+" or "-".
        window_size (int): The desired size of the window.

    Returns:
        tuple[int, int]: The calculated start and end points of the window.

    Raises:
        AssertionError: If the resulting window size doesn't match the
            specified `window_size`.

    Example:
        >>> calculate_window_undersized(10, 50, "+", 20)
        (10, 30)

        >>> calculate_window_undersized(10, 50, "-", 20)
        (30, 50)

    """
    # select from start of gene, since this is such a strong signal for function
    if strand == "+":
        start_window = start
        end_window = start + window_size
    elif strand == "-":
        start_window = end - window_size
        end_window = end
    assert (
        end_window - start_window
    ) == window_size, (
        f"Window sizing is incorrect. Window is larger than {window_size}bp"
    )
    return start_window, end_window


def calculate_window_bounds(
    start: int, end: int, strand: str, window_size: int, chromosome_length: int
) -> tuple[int, int]:
    """
    Calculate the window bounds for genomic sequences.

    This function calculates window bounds for genomic sequences while considering
    strand direction. It ensures the window does not exceed the chromosome length.

    Args:
        start (int): The start point of the sequence on the genome.
        end (int): The end point of the sequence on the genome.
        strand (str): The strand of the gene, either "+" or "-".
        window_size (int): Desired window size.
        chromosome_length (int): Length of the chromosome.

    Returns:
        tuple[int, int]: The calculated start and end points of the window.

    Raises:
        ValueError: If the end position is out of bounds of the chromosome,
            if the start position is greater than or equal to the end position,
            or if the window size is greater than the chromosome length.

    Examples:
        >>> calculate_window_bounds(0, 20, "+", 40, 100)
        (0, 40)
        >>> calculate_window_bounds(5, 25, "+", 50, 100)
        (0, 50)
        >>> calculate_window_bounds(75, 95, "+", 50, 100)
        (50, 100)
        >>> calculate_window_bounds(0, 20, "-", 40, 100)
        (0, 40)
    """
    if end > chromosome_length:
        raise ValueError("End position is out of bounds of chromosome")
    if start >= end:
        raise ValueError("Start position must be less than end position")
    if window_size > chromosome_length:
        raise ValueError("Window size should never be greater than chromosome length")

    seq_length = end - start
    if window_size < seq_length:
        # log info that the window size is smaller than the sequence
        log.info(
            f"Window size {window_size} is smaller than sequence length {end - start}."
        )
        start_window, end_window = calculate_window_undersized(
            start, end, strand, window_size
        )
        return start_window, end_window

    flank_seq_length = (window_size - seq_length) // 2
    start_window = start - flank_seq_length
    end_window = end + flank_seq_length

    # Case where start of window is out of bounds
    if start_window < 0:
        start_window = 0  # Set start of window to beginning of sequence
        end_window = window_size

    # Case where end of window is out of bounds
    if end_window > chromosome_length:
        end_window = chromosome_length  # Set end of window to end of sequence
        start_window = chromosome_length - window_size

    # Edge case if the adjusted window does not match the window size
    # and is up against chromosome ends
    if (
        abs((end_window - start_window) - window_size) == 1
        and end_window == chromosome_length
    ):
        start_window -= 1
    elif abs((end_window - start_window) - window_size) == 1 and start_window == 0:
        end_window += 1
    # Select more of the upstream
    elif abs((end_window - start_window) - window_size) == 1 and strand == "+":
        start_window -= 1
    elif abs((end_window - start_window) - window_size) == 1 and strand == "-":
        end_window += 1
    assert start_window <= start, "Start window must be leq start."
    assert end_window >= end, "End window must be geq end."
    return start_window, end_window


def calculate_window_undersized_symmetric(
    start: int, end: int, window_size: int
) -> tuple[int, int]:
    """
    Calculate symmetric window bounds for genomic sequences.

    This function calculates symmetric window bounds for genomic sequences and
    ensures that the window size is valid and the start and end are not equal.

    Args:
        start (int): The start point of the sequence on the genome.
        end (int): The end point of the sequence on the genome.
        window_size (int): Desired window size.

    Returns:
        tuple[int, int]: The calculated start and end points of the window.

    Raises:
        ValueError: If the start and end positions are the same, or if the
            window size is less than 2.

    Examples:
        >>> calculate_window_undersized_symmetric(10, 20, 4)
        (13, 17)
        >>> calculate_window_undersized_symmetric(10, 20, 5)
        (13, 17)

    Note:
        For odd window sizes, the result will be adjusted to keep the window
        symmetric around the middle of the start and end points.
    """
    if start == end:
        raise ValueError("Start and end positions are the same")
    if window_size < 2:
        raise ValueError("Window size must be at least 2")
    # find the middle
    middle = (start + end) // 2
    # take half above and half below
    flank_size = window_size // 2
    start_window = middle - flank_size
    end_window = middle + flank_size
    # Minus one for odd.
    assert (
        end_window - start_window
    ) >= window_size - 1, (
        f"Window sizing is incorrect. Window is larger than {window_size}bp"
    )
    return start_window, end_window


def calculate_window_bounds_symmetric(
    start: int, end: int, window_size: int, chromosome_length: int
) -> tuple[int, int]:
    r"""
    Calculate symmetric window bounds considering chromosome limits.

    This function calculates symmetric window bounds for genomic sequences, taking
    the chromosome length into account to avoid exceeding it, and also ensuring
    that the window size, start and end positions are valid within these limits.

    Args:
        start (int): The start point of the sequence on the genome.
        end (int): The end point of the sequence on the genome.
        window_size (int): The desired window size.
        chromosome_length (int): The total length of the chromosome.

    Returns:
        tuple[int, int]: The calculated start and end points of the symmetric window.

    Raises:
        ValueError: If the end position is out of chromosome bounds, start position
            is not less than end position, or window size is greater than chromosome
            length.

    Examples:
        >>> calculate_window_bounds_symmetric(5, 15, 30, 100)
        (0, 20)
        >>> calculate_window_bounds_symmetric(80, 95, 30, 100)
        (75, 100)
        >>> calculate_window_bounds_symmetric(45, 55, 30, 100)
        (35, 65)

    Note:
        The function also asserts that the calculated window is within the desired
        limits, symmetric around the middle of the sequence, and the start and end
        of the window are appropriately bounded within the sequence limits.
    """

    if end > chromosome_length:
        raise ValueError("End position is out of bounds of chromosome")
    if start >= end:
        raise ValueError("Start position must be less than end position")
    if window_size > chromosome_length:
        raise ValueError("Window size should never be greater than chromosome length")

    if window_size < end - start:
        start_window, end_window = calculate_window_undersized_symmetric(
            start, end, window_size
        )
        return start_window, end_window

    seq_length = end - start
    # Find limiting window size
    flank_size = (window_size - seq_length) // 2
    start_flank_pos = start - flank_size
    end_flank_pos = end + flank_size
    if start_flank_pos >= 0:
        start_window = start_flank_pos
    else:
        start_window = 0
    if end_flank_pos <= chromosome_length:
        end_window = end_flank_pos
    else:
        end_window = chromosome_length

    # Based on limiting window size take symmetric window
    sym_flank_len = min((start - start_window), (end_window - end))
    start_window = start - sym_flank_len
    end_window = end + sym_flank_len

    assert (
        end_window - start_window
    ) <= window_size, (
        f"Window sizing is incorrect. Window is larger than {window_size}bp"
    )
    assert (
        end_window - start_window
    ) >= end - start, (
        f"Window sizing is incorrect. Window is smaller than sequence {end - start}bp"
    )
    assert (start - start_window) == (end_window - end), "sequences are not symmetric"
    assert start_window <= start, "Start window must be leq start."
    assert end_window >= end, "End window must be geq end."

    return start_window, end_window


class CodonFrequency(SortedDict):
    def __repr__(self):
        if len(self) != 64:
            return "Invalid CodonFrequency: Expected 64 codons"

        sum_freq = sum(self.values())
        if not 0.9999 <= sum_freq <= 1.0001:  # Allowing a small deviation
            return (
                f"Invalid CodonFrequency: Frequencies do not sum to 1 "
                f"(sum={sum_freq})"
            )

        # Sort by frequency in descending order and get the 3 most frequent codons
        most_frequent_codons = sorted(self.items(), key=lambda x: x[1], reverse=True)[
            :3
        ]

        # Round the frequencies to the 4th decimal
        rounded_codons = [
            (codon, round(freq, 4)) for codon, freq in most_frequent_codons
        ]

        return (
            f"CodonFrequency(size={len(self)}, "
            f"most_frequent_codons={rounded_codons}...)"
        )


def compute_codon_frequency(cds_str: str) -> CodonFrequency:
    nucleotides = ["A", "T", "G", "C"]
    all_codons = ["".join(codon) for codon in product(nucleotides, repeat=3)]
    if len(cds_str) % 3 != 0 or not set(cds_str).issubset(set(nucleotides)):
        raise ValueError(
            "Invalid CDS string; length must be a multiple of 3 and only contain A, T, G, C."
        )

    codon_counts = defaultdict(int, {codon: 0 for codon in all_codons})

    for i in range(0, len(cds_str), 3):
        codon = cds_str[i : i + 3]
        codon_counts[codon] += 1

    total_codons = len(cds_str) // 3

    # Convert counts to frequency and create a CodonFrequency object
    codon_frequency = CodonFrequency(
        {codon: count / total_codons for codon, count in codon_counts.items()}
    )

    return codon_frequency


class ParsedGenome(ModelStrictArbitrary):
    gene_set: GeneSet

    @field_validator("gene_set")
    def validate_gene_set(cls, value):
        if not isinstance(value, GeneSet):
            raise ValueError(f"gene_set must be a GeneSet, got {type(value).__name__}")
        return value



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import nn
from torch_geometric.nn import GATConv
from torch_geometric.nn import GCNConv as GCN
from torch_scatter import scatter_add

from torchcell.models import DeepSet


class GraphAttention(nn.Module):
    def __init__(
        self,
        input_dim: int,
        node_layers: list[int],
        set_layers: list[int],
        hidden_channels: int,
        num_layers: int,
        out_channels: int | None = None,
        dropout_prob: float = 0.2,
        norm: str = "batch",
        activation: str = "relu",
        skip_node: bool = False,
        skip_set: bool = False,
        skip_mp: bool = False,
        **kwargs,
    ):
        super().__init__()

        self.deepset = DeepSet(
            input_dim=input_dim,
            node_layers=node_layers,
            set_layers=[],
            dropout_prob=dropout_prob,
            norm=norm,
            activation=activation,
            skip_node=skip_node,
            skip_set=skip_set,
        )
        self.skip_mp = skip_mp
        self.gat_layers = nn.ModuleList()

        if node_layers:
            in_dim = node_layers[-1]
        else:
            in_dim = input_dim

        for i in range(num_layers):
            out_dim = (
                hidden_channels
                if i < num_layers - 1 or out_channels is None
                else out_channels
            )
            # Set to more stable v2
            self.gat_layers.append(GATConv(in_dim, out_dim, v2=True))
            in_dim = out_dim

        # Set layers after GAT
        set_layers.insert(0, in_dim)
        self.set_layers = nn.ModuleList(
            [
                nn.Linear(set_layers[i], set_layers[i + 1])
                for i in range(len(set_layers) - 1)
            ]
        )

    def forward(self, x, batch, edge_index):
        # Process node features
        x_node = self.deepset.node_layers_forward(x)

        # Message passing with GAT
        for layer in self.gat_layers:
            out_node = layer(x_node, edge_index)
            if self.skip_mp and x_node.shape[-1] == out_node.shape[-1]:
                out_node = out_node + x_node  # Skip connection
            x_node = out_node

        x_set = scatter_add(x_node, batch, dim=0)

        # Process aggregated set features
        for i, layer in enumerate(self.set_layers):
            out_set = layer(x_set)
            if self.deepset.skip_set and x_set.shape[-1] == out_set.shape[-1]:
                out_set = out_set + x_set
            x_set = out_set

        return x_node, x_set


def main():
    torch.autograd.set_detect_anomaly(True)

    # Model configuration
    input_dim = 10
    node_layers = [16, 16]
    set_layers = [8, 8]

    model = GraphAttention(
        input_dim,
        node_layers,
        set_layers,
        hidden_channels=16,
        num_layers=2,
        norm="batch",
        activation="gelu",
        skip_node=True,
        skip_set=True,
        skip_mp=True,
    )

    # Dummy data
    x = torch.rand(100, input_dim)
    batch = torch.cat([torch.full((20,), i, dtype=torch.long) for i in range(5)])
    edge_index = torch.stack([torch.arange(100), torch.arange(100)], dim=0)

    # Forward pass
    x_nodes, x_set = model(x, batch, edge_index)
    print(x_set.shape)
    print(x_nodes.shape)

    # Dummy target tensor
    target = torch.rand(5, set_layers[-1])

    # Mean squared error loss
    criterion = nn.MSELoss()
    loss = criterion(x_set, target)
    print("Loss:", loss.item())

    # Backpropagation
    model.zero_grad()
    loss.backward()
    print("Gradients computed successfully!")
    print(model)


if __name__ == "__main__":
    main()



import datetime
import os
import os.path as osp
import pickle
from abc import ABC, abstractmethod

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import networkx as nx
from goatools.obo_parser import GODag
from sortedcontainers import SortedSet
from torch_geometric.data import download_url

from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome

import torchcell

style_file_path = osp.join(osp.dirname(torchcell.__file__), 'torchcell.mplstyle')
plt.style.use(style_file_path)

class GoPlot(ABC):
    @abstractmethod
    def plot(self):
        raise NotImplementedError("Subclasses must implement plot() method.")

    def save(self, as_pdf=False, as_pickle=False):
        current_dir = osp.dirname(osp.abspath(__file__))
        timestamp = datetime.datetime.now().strftime("%Y.%m.%d.%H.%M.%S")
        file_path = osp.join(
            ".notes/assets/images",
            osp.relpath(current_dir),
            f"{self.__class__.__name__}-{timestamp}",
        )
        dir_name = osp.dirname(file_path)
        os.makedirs(dir_name, exist_ok=True)
        plt.savefig(f"{file_path}.png", format="png", bbox_inches="tight")
        if as_pdf:
            plt.savefig(f"{file_path}.pdf", format="pdf", bbox_inches="tight")
        if as_pickle:
            with open(f"{file_path}.pkl", "wb") as f:
                pickle.dump(plt.gcf(), f)


class GOTermGraph:
    def __init__(self, gene_set, add_fantasy_root=False):
        self.gene_set = gene_set
        self.graph = nx.DiGraph()

        data_dir = "data/go"
        obo_path = "data/go/go.obo"
        if not osp.exists(obo_path):
            os.makedirs(data_dir, exist_ok=True)
            download_url("http://current.geneontology.org/ontology/go.obo", data_dir)
            # os.rename(f"{data_dir}/go.obo", obo_path)

        self.godag = GODag(obo_path)

        for term in gene_set:
            # Add a check here
            if term in self.godag:
                self.graph.add_node(term, category=self.godag[term].namespace)
                parents = self.getAllParents(term)
                self.graph.add_edges_from([(term, parent) for parent in parents])
            else:
                print(f"Warning: {term} not found in GO DAG.")

        if add_fantasy_root:
            self.root = "FantasyRoot"
            self.graph.add_node(self.root, category="root")
            for node in self.graph.nodes:
                if self.graph.out_degree(node) == 0 and node != self.root:
                    self.graph.add_edge(node, self.root)
        else:
            self.root = None

    def getAllParents(self, term):
        term_obj = self.godag[term]
        return term_obj.get_all_parents()

    def removeTerm(self, term):
        if term in self.graph:
            self.graph.remove_node(term)


class SimpleTreePlot(GoPlot):
    def __init__(self, go_graph):
        self.go_graph = go_graph

    def plot(self):
        if self.go_graph.root == "FantasyRoot":
            circular_pos = nx.circular_layout(self.go_graph.graph)
            circular_pos[self.go_graph.root] = (0, 0)
        else:
            circular_pos = nx.circular_layout(self.go_graph.graph)

        color_map = []
        for node in self.go_graph.graph.nodes(data=True):
            category = node[1].get("category", "")
            if category == "biological_process":
                color_map.append("green")
            elif category == "cellular_component":
                color_map.append("red")
            elif category == "molecular_function":
                color_map.append("blue")
            elif category == "root":
                color_map.append("purple")
            else:
                color_map.append("lightgray")

        nx.draw(
            self.go_graph.graph,
            pos=circular_pos,
            with_labels=True,
            node_color=color_map,
            arrows=True,
        )

        # Add Legend
        legend_labels = [
            mpatches.Patch(color="green", label="Biological Process"),
            mpatches.Patch(color="red", label="Cellular Component"),
            mpatches.Patch(color="blue", label="Molecular Function"),
            mpatches.Patch(color="purple", label="Fantasy Root"),
        ]
        plt.legend(handles=legend_labels)

    def show(self):
        plt.show()


def main():
    import random

    genome = SCerevisiaeGenome()

    # Take a random sample for testing
    random.seed(0)
    keys_list = list(genome.go_genes.keys())
    sampled_keys = random.sample(keys_list, 10)
    sampled_dict = {key: genome.go_genes[key] for key in sampled_keys}

    go_graph = GOTermGraph(gene_set=sampled_dict, add_fantasy_root=True)
    plotter = SimpleTreePlot(go_graph)
    plotter.plot()
    plotter.show()


if __name__ == "__main__":
    main()



# torchcell/__version__.py
# [[torchcell.__version__]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/__version__.py
# Test file: tests/torchcell/test___version__.py

__version__ = "0.1.17"  # Linked to semantic-release in toml



# torchcell/models/fungal_up_down_transformer.py
# [[torchcell.models.fungal_up_down_transformer]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/fungal_up_down_transformer.py
# Test file: torchcell/models/test_fungal_up_down_transformer.py
import os
import os.path as osp

import torch
from transformers import AutoModelForMaskedLM, AutoTokenizer

from torchcell.models.llm import NucleotideModel


class FungalUpDownTransformer(NucleotideModel):
    r"""
    The FungalUpDownTransformer embeds DNA sequences
    upstream the CDS and downstream the CDS.
    It was trained on 1003 bp upstream sequences (including the start codon) and
    300 bp downstream sequences (including the stop codon).
    It handles both upstream and downstream sequences.
    valid models include :obj:`["downstream_species_lm", "upstream_species_lm"]`.
    The original paper discusses two other models
    (:obj:`["downstream_agnostic_lm"`, :obj:`"upstream_agnostic_lm"]`),
    but they are not supported here.

    Attributes:
        VALID_MODEL_NAMES (List[str]): Valid names for the model
    """

    VALID_MODEL_NAMES = ["downstream_species_lm", "upstream_species_lm"]

    def __init__(
        self, model_name: str = None, target_layer: int | tuple[int, int] = (8,)
    ):
        """
        Initializes a new instance of the FungalUpDownTransformer.

        Args:
            model_name (str, optional): The name of the model to be loaded.
            target_layer (int | tuple[int, int], optional):
            The layer(s) of the model to be used for embeddings.
            If tuple, embeddings are averaged over the range of layers.
            Defaults to (8,).
        """
        self.target_layer = target_layer
        self.tokenizer = None
        self.model = None
        self.model_name = model_name
        self.hugging_model_dir = "gagneurlab/SpeciesLM"
        self.load_model()

    def _check_and_download_model(self) -> None:
        r"""
        Verifies whether the model is downloaded, if not, downloads the model.
        """
        # Define the model name
        model_path = osp.join(self.hugging_model_dir, self.model_name)

        # Define the directory where you want the model to be saved
        script_dir = osp.dirname(osp.realpath(__file__))
        target_directory = osp.join(
            script_dir, "pretrained_LLM", "fungal_up_down_transformer"
        )

        # Create the target directory if it doesn't exist
        if not osp.exists(target_directory):
            os.makedirs(target_directory)  # pragma: no cover

        model_directory = osp.join(target_directory, model_path)

        # Check if the model has already been downloaded
        if osp.exists(model_directory):
            print(f"{model_directory} model already downloaded.")
        else:
            print(f"Downloading {self.model_name} model to {model_directory}...")
            # tokenizer
            AutoTokenizer.from_pretrained(
                self.hugging_model_dir,
                revision=self.model_name,
                cache_dir=target_directory,
            )
            # model
            AutoModelForMaskedLM.from_pretrained(
                self.hugging_model_dir,
                revision=self.model_name,
                cache_dir=target_directory,
            )
            print("Download finished.")

    def load_model(self) -> None:
        r"""
        Loads the required model and tokenizer for embedding sequences.
        """
        self._check_and_download_model()
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.hugging_model_dir, revision=self.model_name
        )
        self.model = AutoModelForMaskedLM.from_pretrained(
            self.hugging_model_dir, revision=self.model_name
        )
        self.model.eval()

    @property
    def max_sequence_size(self) -> int:
        r"""
        Returns the maximum permissible sequence size for the loaded model.

        Returns:
            int: The maximum permissible sequence size.
        """
        if self.model_name.split("_")[0] == "upstream":
            return 1003
        elif self.model_name.split("_")[0] == "downstream":
            return 300

    @staticmethod
    def _pad_sequence(
        tokenized_data: dict[str, torch.tensor], mean_embedding: bool
    ) -> tuple[dict[str, torch.Tensor], int]:
        r"""
        Pads the tokenized sequence to meet the desired length.

        Args:
            tokenized_data (dict[str, torch.tensor]): The tokenized sequence data.
            mean_embedding (bool): Specifies whether to compute the mean embedding.

        Returns:
            tuple[dict[str, torch.Tensor], int]: The padded tokenized data and
             the pad length.

        Raises:
            AssertionError: If not using mean embedding and sequence length is
             below 1001 bp.
        """
        assert (
            mean_embedding
        ), "sequences must be 1003 bp if not using meaning embedding"
        # Find out how much we need to pad
        pad_length = 1001 - tokenized_data["input_ids"].shape[-1]
        pad_start = 2
        pad_end = pad_length + pad_start
        # Padding input_ids, token_type_ids, and attention_mask

        tokenized_data["input_ids"] = torch.cat(
            [
                tokenized_data["input_ids"][0][:2],
                torch.zeros((pad_length), dtype=torch.long),
                tokenized_data["input_ids"][0][2:],
            ]
        ).unsqueeze(0)
        tokenized_data["token_type_ids"] = torch.cat(
            [
                tokenized_data["token_type_ids"][0][:2],
                torch.zeros((pad_length), dtype=torch.long),
                tokenized_data["token_type_ids"][0][2:],
            ]
        ).unsqueeze(0)
        tokenized_data["attention_mask"] = torch.cat(
            [
                tokenized_data["attention_mask"][0][:2],
                torch.zeros((pad_length), dtype=torch.long),
                tokenized_data["attention_mask"][0][2:],
            ]
        ).unsqueeze(0)
        return tokenized_data, pad_start, pad_end

    def embed(
        self,
        sequences: list[str],
        mean_embedding: bool = True,
        proxy_species: str = "candida_glabrata",
    ) -> torch.Tensor:
        r"""
        Embeds the provided sequences using the loaded transformer model
        """
        embeddings = []

        def kmers_stride1(seq, k=6):
            return [seq[i : i + k] for i in range(0, len(seq) - k + 1)]

        for sequence in sequences:
            sequence_length = len(sequence)
            desired_length = self.max_sequence_size

            if (
                self.model_name.startswith("upstream")
                and sequence_length > desired_length
            ):
                raise ValueError(
                    f"Seq len for {self.model_name} must be <= {desired_length}."
                    f" Provided: {sequence_length}"
                )

            if self.model_name.startswith("downstream"):
                if sequence_length < 11:
                    raise ValueError(
                        f"Seq len for {self.model_name} must be >  11."
                        f" Provided: {sequence_length}"
                    )
                elif sequence_length > desired_length:
                    raise ValueError(
                        f"Seq len for {self.model_name} must be <= {desired_length}."
                        f" Provided: {sequence_length}"
                    )

            # Tokenizing the sequence with proxy_species and kmers_stride1
            tokenized_data = self.tokenizer(
                proxy_species + " " + " ".join(kmers_stride1(sequence)),
                return_tensors="pt",
            )

            # Keep track of the original seq len if slice needed later
            input_ids_len = tokenized_data["input_ids"].shape[-1]

            if self.model_name.startswith("upstream") and input_ids_len < 1001:
                tokenized_data, pad_start, pad_end = self._pad_sequence(
                    tokenized_data, mean_embedding
                )

            with torch.no_grad():
                outputs = self.model(**tokenized_data, output_hidden_states=True)
                hidden_states = outputs.hidden_states

            # Handling different cases for target_layer as in the original work
            if isinstance(self.target_layer, int):
                embedding = hidden_states[self.target_layer][0]
            elif len(self.target_layer) == 1:
                embedding = torch.mean(
                    torch.stack(hidden_states[self.target_layer[0] :]), dim=0
                )[0]
            else:
                if self.target_layer[1] > len(hidden_states):
                    raise ValueError(
                        f"Target layer {self.target_layer[1]} is out of range."
                        f" Max layer is {len(hidden_states)}."
                    )
                embedding = torch.mean(
                    torch.stack(
                        hidden_states[self.target_layer[0] : self.target_layer[1] + 1]
                    ),
                    dim=0,
                )[0]

            if self.model_name.startswith("upstream") and input_ids_len < 1001:
                embedding = torch.concat(
                    [embedding[:pad_start, :], embedding[pad_end:, :]]
                )

            if mean_embedding:
                embedding = embedding.mean(dim=0).cpu()

            embeddings.append(embedding)

        embeddings_tensor = (
            torch.stack(embeddings)
            if not mean_embedding
            else torch.cat(embeddings).view(len(sequences), -1)
        )
        return embeddings_tensor


if __name__ == "__main__":
    model = FungalUpDownTransformer(model_name="upstream_species_lm")
    model.embed(["A" * 500 + "ATG"], mean_embedding=True)



# torchcell/models/protT5.py
# [[torchcell.models.protT5]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/protT5.py
# Test file: torchcell/models/test_protT5.py

import os
import os.path as osp
import re

import torch
from transformers import T5EncoderModel, T5Tokenizer

from torchcell.models.llm import PeptideModel


class ProtT5(PeptideModel):
    VALID_MODEL_NAMES = ["prot_t5_xl_uniref50"]

    def __init__(self, model_name: str):
        self.tokenizer = None
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model_name = osp.join("Rostlab", model_name)
        super().__init__(self.model_name)

    @property
    def max_sequence_size(self):
        return 40000

    @staticmethod
    def _check_and_download_model(model_name: str):
        # Define the directory where you want the model to be saved
        script_dir = osp.dirname(osp.realpath(__file__))
        target_directory = osp.join(script_dir, "pretrained_LLM", "ProtT5")

        # Create the target directory if it doesn't exist
        if not osp.exists(target_directory):
            os.makedirs(target_directory)

        model_directory = osp.join(target_directory, model_name)

        # Check if the model has already been downloaded
        if osp.exists(model_directory):
            print(f"{model_name} model already downloaded.")
        else:
            print(f"Downloading {model_name} model to {model_directory}...")
            # tokenizer
            T5Tokenizer.from_pretrained(
                model_name, cache_dir=target_directory, legacy=True
            )
            # model
            T5EncoderModel.from_pretrained(model_name, cache_dir=target_directory)
            print("Download finished.")

    def _prepare_sequence(self, sequence: str) -> str:
        # Replace all rare/ambiguous amino acids by X
        # Introduce white-space between all amino acids
        return " ".join(list(re.sub(r"[UZOB]", "X", sequence)))

    def load_model(self, model_name: str):
        # Check and download the model if necessary
        self._check_and_download_model(model_name)

        # Load the tokenizer and the model
        self.tokenizer = T5Tokenizer.from_pretrained(model_name, do_lower_case=False)
        self.model = T5EncoderModel.from_pretrained(model_name).to(self.device)

        # Set to half-precision if on GPU, full-precision if on CPU
        self.model.full() if self.device == "cpu" else self.model.half()

    def embed(self, sequences: list[str], mean_embedding: bool = False) -> torch.Tensor:
        # Pre-process sequences
        sequences = [self._prepare_sequence(seq) for seq in sequences]

        # Tokenize sequences and pad up to the longest sequence in the batch
        ids = self.tokenizer(
            sequences, add_special_tokens=True, padding="longest", return_tensors="pt"
        )
        input_ids = ids["input_ids"].to(self.device)
        attention_mask = ids["attention_mask"].to(self.device)

        # Generate embeddings
        with torch.no_grad():
            embedding_repr = self.model(
                input_ids=input_ids, attention_mask=attention_mask
            )

        embeddings = embedding_repr.last_hidden_state

        if mean_embedding:
            # Compute mean embeddings per sequence
            embeddings = embeddings.mean(dim=1)

        return embeddings


if __name__ == "__main__":
    model = ProtT5(model_name="prot_t5_xl_uniref50")
    sample_sequences = ["P" * 4911]
    embeddings = model.embed(sample_sequences, mean_embedding=True)
    print(embeddings.shape)



# torchcell/ontology/tc_ontology.py
# [[torchcell.ontology.tc_ontology]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/ontology/tc_ontology.py
# Test file: torchcell/ontology/test_tc_ontology.py



# torchcell/trainers/dcell_regression_slim.py
# [[torchcell.trainers.dcell_regression_slim]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/trainers/dcell_regression_slim.py
# Test file: torchcell/trainers/test_dcell_regression_slim.py

import math

import lightning as L
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from torch_geometric.data import Batch, Data
from torchmetrics import (
    MeanAbsoluteError,
    MeanSquaredError,
    MetricCollection,
    PearsonCorrCoef,
    SpearmanCorrCoef,
)

# from torchmetrics.regression import PearsonCorrCoef, SpearmanCorrCoef
from tqdm import tqdm

import wandb
from torchcell.losses import DCellLoss, WeightedMSELoss
from torchcell.viz import fitness, genetic_interaction_score

import torchcell

style_file_path = osp.join(osp.dirname(torchcell.__file__), 'torchcell.mplstyle')
plt.style.use(style_file_path)

class DCellRegressionSlimTask(L.LightningModule):
    """LightningModule for training models on graph-based regression datasets."""

    def __init__(
        self,
        models: dict[str, nn.Module],
        target: str,
        boxplot_every_n_epochs: int = 10,
        learning_rate: float = 1e-3,
        weight_decay: float = 1e-5,
        batch_size: int = None,
        alpha: float = 0.3,
        **kwargs,
    ):
        super().__init__()
        # models
        self.models = models

        for key, value in models.items():
            setattr(self, key, value)

        # target for training
        self.target = target

        # Lightning settings, doing this for WT embedding
        self.automatic_optimization = False

        self.x_name = "x"
        self.x_batch_name = "batch"

        self.loss = DCellLoss(alpha)

        # optimizer
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay

        # batch_size
        self.batch_size = batch_size

        self.train_metrics = MetricCollection(
            {
                "RMSE": MeanSquaredError(squared=False),
                "MSE": MeanSquaredError(squared=True),
                "MAE": MeanAbsoluteError(),
                "Pearson": PearsonCorrCoef(),
                "Spearman": SpearmanCorrCoef(),
            },
            prefix="train_",
        )
        self.val_metrics = self.train_metrics.clone(prefix="val_")
        self.test_metrics = self.train_metrics.clone(prefix="test_")
        self.train_metrics_root = MetricCollection(
            {
                "RMSE": MeanSquaredError(squared=False),
                "MSE": MeanSquaredError(squared=True),
                "MAE": MeanAbsoluteError(),
                "Pearson": PearsonCorrCoef(),
                "Spearman": SpearmanCorrCoef(),
            },
            prefix="train_root_",
        )
        self.val_metrics_root = self.train_metrics_root.clone(prefix="val_root_")
        self.test_metrics_root = self.train_metrics_root.clone(prefix="test_root_")

        # Used in end for whisker plot
        self.boxplot_every_n_epochs = boxplot_every_n_epochs

        # wandb model artifact logging
        self.last_logged_best_step = None

    def setup(self, stage=None):
        for model in self.models.values():
            model.to(self.device)

    def forward(self, batch):
        # Implement the forward pass
        dcell_subsystem_output = self.dcell(batch)
        dcell_linear_output = self.dcell_linear(dcell_subsystem_output)
        # if dcell_linear_output.size()[-1] == 1:
        #     dcell_linear_output = dcell_linear_output.squeeze(-1)
        return dcell_linear_output

    def on_train_start(self):
        # Calculate the model size (number of parameters)
        parameter_size = sum(p.numel() for p in self.parameters())
        # Log it using wandb
        self.log(
            "model/parameters_size", torch.tensor(parameter_size, dtype=torch.float32)
        )

    def training_step(self, batch, batch_idx):
        y_hat = self(batch)
        y = batch.fitness
        opt = self.optimizers()
        opt.zero_grad()
        loss = self.loss(y_hat, y, self.dcell.parameters())

        self.manual_backward(loss)  # error on this line
        opt.step()
        opt.zero_grad()
        # logging
        batch_size = batch.batch[-1].item() + 1
        # Flatten
        y_hat_root = y_hat["GO:ROOT"].squeeze(1)
        y_hat_stacked = torch.stack([v.squeeze() for v in y_hat.values()])
        y_hat_subsystems = y_hat_stacked.mean(0)
        # Log
        self.log("train_loss", loss, batch_size=batch_size, sync_dist=True)
        self.train_metrics(y_hat_subsystems, y)
        self.train_metrics_root(y_hat_root, y)
        return loss

    def on_train_epoch_end(self):
        self.log_dict(self.train_metrics.compute(), sync_dist=True)
        self.log_dict(self.train_metrics_root.compute(), sync_dist=True)
        self.train_metrics.reset()
        self.train_metrics_root.reset()
        pass

    def validation_step(self, batch, batch_idx):
        # Extract the batch vector
        y_hat = self(batch)
        y = batch.fitness
        loss = self.loss(y_hat, y, self.dcell.parameters())
        batch_size = batch.batch[-1].item() + 1
        self.log("val_loss", loss, batch_size=batch_size, sync_dist=True)
        # Flatten
        y_hat_root = y_hat["GO:ROOT"].squeeze(1)
        y_hat_stacked = torch.stack([v.squeeze() for v in y_hat.values()])
        y_hat_subsystems = y_hat_stacked.mean(0)
        # Log
        self.val_metrics(y_hat_subsystems, y)
        self.val_metrics_root(y_hat_root, y)

    def on_validation_epoch_end(self):
        self.log_dict(self.val_metrics.compute(), sync_dist=True)
        self.log_dict(self.val_metrics_root.compute(), sync_dist=True)
        self.val_metrics.reset()
        self.val_metrics_root.reset()

        # Stop tracing memory allocations
        current_global_step = self.global_step
        if (
            self.trainer.checkpoint_callback.best_model_path
            and current_global_step != self.last_logged_best_step
        ):
            # Save model as a W&B artifact
            artifact = wandb.Artifact(
                name=f"model-global_step-{current_global_step}",
                type="model",
                description=f"Model on validation epoch end step - {current_global_step}",
                metadata=dict(self.hparams),
            )
            artifact.add_file(self.trainer.checkpoint_callback.best_model_path)
            wandb.log_artifact(artifact)
            self.last_logged_best_step = (
                current_global_step  # update the last logged step
            )

    def test_step(self, batch, batch_idx):
        y_hat = self(batch)
        y = batch.fitness
        loss = self.loss(y_hat, y, self.dcell.parameters())
        batch_size = batch.batch[-1].item() + 1
        # Flatten
        y_hat_root = y_hat["GO:ROOT"].squeeze(1)
        y_hat_stacked = torch.stack([v.squeeze() for v in y_hat.values()])
        y_hat_subsystems = y_hat_stacked.mean(0)
        #
        self.log("test_loss", loss, batch_size=batch_size, sync_dist=True)
        self.test_metrics(y_hat_subsystems, y)
        self.test_metrics_root(y_hat_root, y)

    def on_test_epoch_end(self):
        self.log_dict(self.test_metrics.compute(), sync_dist=True)
        self.test_metrics.reset()

    def configure_optimizers(self):
        params = list(self.models["dcell"].parameters()) + list(
            self.models["dcell_linear"].parameters()
        )
        optimizer = torch.optim.Adam(
            params, lr=self.learning_rate, weight_decay=self.weight_decay
        )
        return optimizer



from typing import Any, Dict

from pydantic import BaseModel, ConstrainedStr, validator


class VarNameStr(ConstrainedStr):
    @classmethod
    def validate(cls, v: str) -> str:
        if not v.isidentifier():
            raise ValueError("String is not a valid Python identifier")
        return v


class ModelIndex(BaseModel):
    __root__: dict[VarNameStr, Any]

    @validator("__root__", pre=True, each_item=True)
    def check_keys(cls, v, field, values, **kwargs):
        if not v.isidentifier():
            raise ValueError(f"Invalid attribute name: {v}")
        return v

    def __getattr__(self, item):
        return self.__root__[item]


# Example usage
try:
    model_index = ModelIndex(
        __root__={
            "dcell": "some_value",  # Replace with your actual object
            "dcell_linear": "another_value",
            "1invalid": "this_will_fail",
        }
    )
except ValueError as e:
    print(e)  # Will raise an error for '1invalid'

# Accessing attributes
print(model_index.dcell)
print(model_index.dcell_linear)



# torchcell/models/species_aware_lm.py
# [[torchcell.models.species_aware_lm]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/species_aware_lm.py
# Test file: torchcell/models/test_species_aware_lm.py


import collections
import itertools
import math
import os
from collections.abc import Mapping

import numpy as np
import pandas as pd
import torch
import tqdm
from datasets import Dataset
from transformers import (
    AutoConfig,
    AutoModelForMaskedLM,
    AutoModelForSequenceClassification,
    AutoTokenizer,
    DataCollatorForLanguageModeling,
    DataCollatorWithPadding,
    Trainer,
)

from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome


def embed_sequence(sequence: str) -> np.ndarray:
    def kmers_stride1(seq, k=6):
        return [seq[i : i + k] for i in range(0, len(seq) - k + 1)]

    # Tokenizing the sequence
    proxy_species = "candida_glabrata"
    tokenized_data = tokenizer(
        proxy_species + " " + " ".join(kmers_stride1(sequence)),
        return_tensors="pt",  # Return PyTorch tensors
    )

    # Running model and getting embeddings
    with torch.no_grad():
        outputs = model(**tokenized_data, output_hidden_states=True)
        hidden_states = outputs.hidden_states

    target_layer = 8  # Use the layer that suits your needs

    if isinstance(target_layer, int):
        embedding = hidden_states[target_layer]
    else:
        # If target_layer is a tuple or list, average over the specified range of layers
        embedding = torch.mean(
            torch.stack(hidden_states[target_layer[0] : target_layer[1] + 1]), dim=0
        )

    # Averaging over tokens to get a single vector and converting it to a NumPy array
    averaged_embedding = embedding.mean(dim=1).cpu().numpy()

    return averaged_embedding


# Initialize tokenizer and model outside the function
tokenizer = AutoTokenizer.from_pretrained(
    "gagneurlab/SpeciesLM", revision="downstream_species_lm"
)
model = AutoModelForMaskedLM.from_pretrained(
    "gagneurlab/SpeciesLM", revision="downstream_species_lm"
)
model.eval()  # Set the model to evaluation mode


def main():
    genome = SCerevisiaeGenome()

    tokenizer = AutoTokenizer.from_pretrained(
        "gagneurlab/SpeciesLM", revision="downstream_species_lm"
    )
    model = AutoModelForMaskedLM.from_pretrained(
        "gagneurlab/SpeciesLM", revision="downstream_species_lm"
    )

    def kmers_stride1(seq, k=6):
        # splits a sequence into overlapping k-mers
        return [seq[i : i + k] for i in range(0, len(seq) - k + 1)]

    def tok_func_species(x, species_proxy, seq_col):
        res = tokenizer(species_proxy + " " + " ".join(kmers_stride1(x[seq_col])))
        return res

    #
    seq_col = (
        "three_prime_seq"  # name of the column in the df that stores the sequences
    )

    proxy_species = "candida_glabrata"  # species token to use
    target_layer = (8,)  # what hidden layers to use for embedding

    #
    tok_func = lambda x: tok_func_species(x, proxy_species, seq_col)

    # I want a function that would all me to run embed(genome[gene].window_three_prime(300, include_stop_codon=True, allow_undersize=True).seq) and it would return the embedded vector

    genes_three_prime = []
    for gene in genome.gene_set:
        genes_three_prime.append(
            genome[gene]
            .window_three_prime(300, include_stop_codon=True, allow_undersize=True)
            .seq
        )
    dataset = pd.DataFrame({seq_col: genes_three_prime})

    ds = Dataset.from_pandas(dataset[[seq_col]])

    tok_ds = ds.map(tok_func, batched=False, num_proc=2)

    rem_tok_ds = tok_ds.remove_columns(seq_col)

    data_collator = DataCollatorForLanguageModeling(tokenizer, mlm=False)
    data_loader = torch.utils.data.DataLoader(
        rem_tok_ds, batch_size=1, collate_fn=data_collator, shuffle=False
    )
    # Running model

    # CHECK needed for running model?
    def count_special_tokens(tokens, tokenizer, where="left"):
        count = 0
        if where == "right":
            tokens = tokens[::-1]
        for pos in range(len(tokens)):
            tok = tokens[pos]
            if tok in tokenizer.all_special_ids:
                count += 1
            else:
                break
        return count

    def embed_on_batch(
        tokenized_data,
        dataset,
        seq_idx,
        special_token_offset,
        target_layer=target_layer,
    ):
        label = dataset.iloc[seq_idx][seq_col]
        label_len = len(label)
        if label_len < 6:
            print("This should not occur")
            return torch.zeros(label_len, label_len, 768)
        else:
            res = tokenized_data["input_ids"].clone()
            # can change later
            device = "cpu"
            res = res.to(device)
            with torch.no_grad():
                embedding = model(res, output_hidden_states=True)["hidden_states"]
        if isinstance(target_layer, int):
            embedding = embedding[target_layer]
        elif len(target_layer) == 1:
            embedding = torch.stack(embedding[target_layer[0] :], axis=0)
            embedding = torch.mean(embedding, axis=0)
        else:
            embedding = torch.stack(
                embedding[target_layer[0] : target_layer[1]], axis=0
            )
            embedding = torch.mean(embedding, axis=0)
        embedding = embedding.detach().cpu().numpy()
        return embedding

    averaged_embeddings = []

    for no_of_index, tokenized_data in tqdm.tqdm(enumerate(data_loader)):
        label = dataset.iloc[no_of_index][seq_col]
        label_len = len(label)

        left_special_tokens = count_special_tokens(
            tokenized_data["input_ids"].numpy()[0], tokenizer, where="left"
        )
        # TODO why are these computed? Are they needed for upstream?
        # right_special_tokens = count_special_tokens(
        #     tokenized_data["input_ids"].numpy()[0], tokenizer, where="right"
        # )
        assert label_len > 11, "Cannot embed sequences < 11"

        hidden_states = embed_on_batch(
            tokenized_data,
            dataset,
            no_of_index,
            special_token_offset=left_special_tokens,
        )
        avg = hidden_states.mean(
            axis=(0, 1)
        )  # breakpoint here, hidden state size is (1, 298, 768)

        averaged_embeddings.append(avg)
        if no_of_index > 10:
            break
    print()


if __name__ == "__main__":
    # main()
    # main()
    # single embed
    # Get genome sequence. Replace the next line with the actual way you are getting the sequence.
    genome = SCerevisiaeGenome()
    genome.drop_chrmt()
    genome.drop_empty_go()
    count = 0
    lens = []
    for gene in genome.gene_set:
        try:
            genome[gene].window_five_prime(
                1003, include_start_codon=True, allow_undersize=False
            )
        except ValueError:
            count += 1
            lens.append(
                len(
                    genome[gene]
                    .window_five_prime(
                        1003, include_start_codon=True, allow_undersize=True
                    )
                    .seq
                )
            )

    genome_sequence = (
        genome["YDR210W"]
        .window_three_prime(300, include_stop_codon=True, allow_undersize=True)
        .seq
    )

    # Get the embedded vector by calling the function
    embedded_vector = embed_sequence(genome_sequence)
    print()



# torchcell/datamodels/schema
# [[torchcell.datamodels.schema]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datamodels/schema
# Test file: tests/torchcell/datamodels/test_schema.py


import json
from typing import List, Union, Optional

from pydantic import BaseModel, Field, field_validator, root_validator
from enum import Enum, auto
from torchcell.datamodels.pydant import ModelStrict


# Genotype
class ReferenceGenome(ModelStrict):
    species: str
    strain: str


class GenePerturbation(ModelStrict):
    systematic_gene_name: str
    perturbed_gene_name: str

    @field_validator("systematic_gene_name", mode="after")
    @classmethod
    def validate_sys_gene_name(cls, v):
        if len(v) < 7 or len(v) > 9:
            raise ValueError("Systematic gene name must be between 7 and 9 characters")
        return v

    @field_validator("perturbed_gene_name", mode="after")
    @classmethod
    def validate_pert_gene_name(cls, v):
        if v.endswith("'"):
            v = v[:-1] + "_prime"
        return v


class DeletionPerturbation(GenePerturbation, ModelStrict):
    description: str = "Deletion via KanMX or NatMX gene replacement"
    perturbation_type: str = "deletion"


class KanMxDeletionPerturbation(DeletionPerturbation, ModelStrict):
    deletion_description: str = "Deletion via KanMX gene replacement."
    deletion_type: str = "KanMX"


class NatMxDeletionPerturbation(DeletionPerturbation, ModelStrict):
    deletion_description: str = "Deletion via NatMX gene replacement."
    deletion_type: str = "NatMX"


class SgaKanMxDeletionPerturbation(KanMxDeletionPerturbation, ModelStrict):
    kan_mx_description: str = (
        "KanMX Deletion Perturbation information specific to SGA experiments."
    )
    strain_id: str = Field(description="'Strain ID' in raw data.")
    kanmx_deletion_type: str = "SGA"


class SgaNatMxDeletionPerturbation(NatMxDeletionPerturbation, ModelStrict):
    nat_mx_description: str = (
        "NatMX Deletion Perturbation information specific to SGA experiments."
    )
    strain_id: str = Field(description="'Strain ID' in raw data.")
    natmx_deletion_type: str = "SGA"

    # @classmethod
    # def _process_perturbation_data(cls, perturbation_data):
    #     if isinstance(perturbation_data, list):
    #         return [cls._create_perturbation_from_dict(p) for p in perturbation_data]
    #     elif isinstance(perturbation_data, dict):
    #         return cls._create_perturbation_from_dict(perturbation_data)
    #     return perturbation_data


class ExpressionRangeMultiplier(ModelStrict):
    min: float = Field(
        ..., description="Minimum range multiplier of gene expression levels"
    )
    max: float = Field(
        ..., description="Maximum range multiplier of gene expression levels"
    )


class DampPerturbation(GenePerturbation, ModelStrict):
    description: str = "4-10 decreased expression via KANmx insertion at the "
    "the 3' UTR of the target gene."
    expression_range: ExpressionRangeMultiplier = Field(
        default=ExpressionRangeMultiplier(min=1 / 10.0, max=1 / 4.0),
        description="Gene expression is decreased by 4-10 fold",
    )
    perturbation_type: str = "damp"


class SgaDampPerturbation(DampPerturbation, ModelStrict):
    damp_description: str = "Damp Perturbation information specific to SGA experiments."
    strain_id: str = Field(description="'Strain ID' in raw data.")
    damp_perturbation_type: str = "SGA"


class TsAllelePerturbation(GenePerturbation, ModelStrict):
    description: str = (
        "Temperature sensitive allele compromised by amino acid substitution."
    )
    # seq: str = "NOT IMPLEMENTED"
    perturbation_type: str = "temperature_sensitive_allele"


class AllelePerturbation(GenePerturbation, ModelStrict):
    description: str = (
        "Allele compromised by amino acid substitution without more generic"
        "phenotypic information specified."
    )
    # seq: str = "NOT IMPLEMENTED"
    perturbation_type: str = "allele"


class SuppressorAllelePerturbation(GenePerturbation, ModelStrict):
    description: str = (
        "suppressor allele that results in higher fitness in the presence"
        "of a perturbation, compared to the fitness of the perturbation alone."
    )
    perturbation_type: str = "suppressor_allele"


class SgaSuppressorAllelePerturbation(SuppressorAllelePerturbation, ModelStrict):
    suppressor_description: str = (
        "Suppressor Allele Perturbation information specific to SGA experiments."
    )
    strain_id: str = Field(description="'Strain ID' in raw data.")
    suppressor_allele_perturbation_type: str = "SGA"


class SgaTsAllelePerturbation(TsAllelePerturbation, ModelStrict):
    ts_allele_description: str = (
        "Ts Allele Perturbation information specific to SGA experiments."
    )
    strain_id: str = Field(description="'Strain ID' in raw data.")
    temperature_sensitive_allele_perturbation_type: str = "SGA"


class SgaAllelePerturbation(AllelePerturbation, ModelStrict):
    allele_description: str = (
        "Ts Allele Perturbation information specific to SGA experiments."
    )
    strain_id: str = Field(description="'Strain ID' in raw data.")
    allele_perturbation_type: str = "SGA"


SgaPerturbationType = Union[
    SgaKanMxDeletionPerturbation,
    SgaNatMxDeletionPerturbation,
    SgaDampPerturbation,
    SgaTsAllelePerturbation,
    SgaSuppressorAllelePerturbation,
    SgaAllelePerturbation,
]

GenePerturbationType = Union[SgaPerturbationType]


class Genotype(ModelStrict):
    perturbations: list[GenePerturbationType] = Field(description="Gene perturbation")

    @field_validator("perturbations", mode="after")
    @classmethod
    def sort_perturbations(cls, perturbations):
        return sorted(
            perturbations,
            key=lambda p: (
                p.systematic_gene_name,
                p.perturbation_type,
                p.perturbed_gene_name,
            ),
        )

    @property
    def systematic_gene_names(self):
        sorted_perturbations = sorted(
            self.perturbations, key=lambda p: p.systematic_gene_name
        )
        return [p.systematic_gene_name for p in sorted_perturbations]

    @property
    def perturbed_gene_names(self):
        sorted_perturbations = sorted(
            self.perturbations, key=lambda p: p.systematic_gene_name
        )
        return [p.perturbed_gene_name for p in sorted_perturbations]

    @property
    def perturbation_types(self):
        sorted_perturbations = sorted(
            self.perturbations, key=lambda p: p.systematic_gene_name
        )
        return [p.perturbation_type for p in sorted_perturbations]

    def __len__(self):
        return len(self.perturbations)

    # we would use set, but need serialization to be a list
    def __eq__(self, other):
        if not isinstance(other, Genotype):
            return NotImplemented

        return set(self.perturbations) == set(other.perturbations)


# Environment
class Media(ModelStrict):
    name: str
    state: str

    @field_validator("state", mode="after")
    @classmethod
    def validate_state(cls, v):
        if v not in ["solid", "liquid", "gas"]:
            raise ValueError('state must be one of "solid", "liquid", or "gas"')
        return v


class Temperature(BaseModel):
    value: float  # Renamed from scalar to value
    unit: str = "Celsius"  # Simplified unit string

    @field_validator("value", mode="after") 
    @classmethod
    def check_temperature(cls, v):
        if v < -273:
            raise ValueError("Temperature cannot be below -273 degrees Celsius")
        return v


class BaseEnvironment(ModelStrict):
    media: Media
    temperature: Temperature


# Phenotype


class BasePhenotype(ModelStrict):
    graph_level: str
    label: str
    label_error: str

    @field_validator("graph_level", mode="after")
    @classmethod
    def validate_level(cls, v):
        levels = {"edge", "node", "subgraph", "global", "metabolism"}

        if v not in levels:
            raise ValueError("level must be one of: edge, node, global, metabolism")

        return v


class FitnessPhenotype(BasePhenotype, ModelStrict):
    fitness: float = Field(description="wt_growth_rate/ko_growth_rate")
    fitness_std: Optional[float] = Field(None, description="fitness standard deviation")


# TODO when we only do BasePhenotype during serialization, we will lose the other information. It might be good to make refs for each phenotype,
class ExperimentReference(ModelStrict):
    reference_genome: ReferenceGenome
    reference_environment: BaseEnvironment
    reference_phenotype: BasePhenotype


class BaseExperiment(ModelStrict):
    genotype: Genotype
    environment: BaseEnvironment
    phenotype: BasePhenotype


class FitnessExperimentReference(ExperimentReference, ModelStrict):
    reference_phenotype: FitnessPhenotype


class FitnessExperiment(BaseExperiment):
    genotype: Union[Genotype, List[Genotype,]]
    phenotype: FitnessPhenotype


if __name__ == "__main__":
    pass



import json

from neo4j import GraphDatabase


def make_fake_data_1():
    some_serialized_data = {"key1": "value1", "key2": 123}  # Example serialized data
    fake_data = {
        "genotype": [
            {"id": "YPL264C", "intervention": "deletion", "id_full": "YPL264C_sn3676"},
            {"id": "YNL051W", "intervention": "deletion", "id_full": "YNL051W_dma3963"},
        ],
        "phenotype": {
            "dmf": 0.9457,
            "dmf_std": 0.0292,
            "genetic_interaction_score": 0.0206,
            "genetic_interaction_p_value": 0.2841,
            "special_fitness": some_serialized_data,  # Serialized data added
        },
        "environment": {"media": "Ypd", "temperature": 30},
    }
    return fake_data


def make_fake_data_2():
    some_serialized_data = {"key1": "value1", "key2": 123}  # Example serialized data
    fake_data = {
        "genotype": [
            {"id": "YFL041W", "intervention": "deletion", "id_full": "YFL041W_sn2437"},
            {"id": "YDR097C", "intervention": "deletion", "id_full": "YDR097C_dma3963"},
        ],
        "phenotype": {
            "dmf": 1.0265,
            "dmf_std": 0.0034,
            "genetic_interaction_score": -0.0045,
            "genetic_interaction_p_value": 0.2112,
            "special_fitness": some_serialized_data,  # Serialized data added
        },
        "environment": {"media": "YPD", "temperature": 30},
    }
    return fake_data


def upload_data_instance(session, data_instance_name, data_instance):
    # Create the main data instance node with dynamic name
    session.run(f"CREATE (di:DataInstance {{name: '{data_instance_name}'}})")

    # Base URI of your ontology
    base_ontology_uri = (
        "https://raw.githubusercontent.com/Mjvolk3/torchcell/main/torchcell.rdf"
    )

    for genotype in data_instance["genotype"]:
        if genotype.get("intervention") == "deletion":
            # Link the Deletion node to the ontology Resource node
            session.run(
                f"""
                MATCH (di:DataInstance {{name: '{data_instance_name}'}})
                MERGE (deletion:Deletion {{id: $id, id_full: $id_full, intervention: $intervention}})
                MERGE (di)-[:HAS_GENOTYPE]->(deletion)
                WITH deletion
                MATCH (ont:Resource {{uri: '{base_ontology_uri}#GeneDeletion'}})
                MERGE (deletion)-[:IS_A]->(ont)
                """,
                {
                    "id": genotype["id"],
                    "id_full": genotype["id_full"],
                    "intervention": genotype["intervention"],
                },
            )


# def upload_data_instance(session, data_instance_name, data_instance):
#     # Create the main data instance node with dynamic name
#     session.run(f"CREATE (di:DataInstance {{name: '{data_instance_name}'}})")

#     # Base URI of your ontology
#     base_ontology_uri = "http://example.org/onto.owl"

#     # Create Genotype nodes and link them to the data instance and ontology
#     for genotype in data_instance["genotype"]:
#         session.run(
#             f"""
#             MATCH (di:DataInstance {{name: '{data_instance_name}'}})
#             MERGE (g:Genotype {{id: $id, intervention: $intervention, id_full: $id_full}})
#             MERGE (di)-[:HAS_GENOTYPE]->(g)
#             MERGE (ont:Resource {{uri: '{base_ontology_uri}#Genotype'}})
#             MERGE (g)-[:IS_A]->(ont)
#             """,
#             genotype,
#         )

# # Create Phenotype nodes and link them to the data instance and ontology
# phenotype = data_instance["phenotype"]
# serialized_data = json.dumps(phenotype["special_fitness"])
# session.run(
#     f"""
#     MATCH (di:DataInstance {{name: '{data_instance_name}'}})
#     MERGE (p:Phenotype {{dmf: $dmf, dmf_std: $dmf_std,
#         genetic_interaction_score: $genetic_interaction_score,
#         genetic_interaction_p_value: $genetic_interaction_p_value,
#         special_fitness: $special_fitness}})
#     MERGE (di)-[:HAS_PHENOTYPE]->(p)
#     MERGE (ont:Resource {{uri: '{base_ontology_uri}#Phenotype'}})
#     MERGE (p)-[:IS_A]->(ont)
#     """,
#     {
#         "dmf": phenotype["dmf"],
#         "dmf_std": phenotype["dmf_std"],
#         "genetic_interaction_score": phenotype["genetic_interaction_score"],
#         "genetic_interaction_p_value": phenotype["genetic_interaction_p_value"],
#         "special_fitness": serialized_data,
#     },
# )

# # Create Environment node and link it to the data instance and ontology
# environment = data_instance["environment"]
# session.run(
#     f"""
#     MATCH (di:DataInstance {{name: '{data_instance_name}'}})
#     MERGE (e:Environment {{media: $media, temperature: $temperature}})
#     MERGE (di)-[:HAS_ENVIRONMENT]->(e)
#     MERGE (ont:Resource {{uri: '{base_ontology_uri}#Environment'}})
#     MERGE (e)-[:IS_A]->(ont)
#     MERGE (m:Media {{name: $media}})
#     MERGE (e)-[:HAS_MEDIA]->(m)
#     MERGE (t:Temperature {{value: $temperature}})
#     MERGE (e)-[:HAS_TEMPERATURE]->(t)
#     """,
#     environment,
# )


def add_data():
    data_0 = make_fake_data_1()
    data_1 = make_fake_data_2()

    uri = "neo4j://localhost:7687"
    username = "neo4j"
    password = "torchcell"

    driver = GraphDatabase.driver(uri, auth=(username, password))
    with driver.session() as session:
        upload_data_instance(session, "dmf_dataset_0", data_0)
        upload_data_instance(session, "dmf_dataset_1", data_1)
    driver.close()


def fetch_data_instance(uri, username, password):
    driver = GraphDatabase.driver(uri, auth=(username, password))
    data_instance = {}

    with driver.session() as session:
        result = session.run(
            "MATCH (di:DataInstance)-[:HAS_GENOTYPE]->(g:Genotype), "
            "(di)-[:HAS_PHENOTYPE]->(p:Phenotype), "
            "(di)-[:HAS_ENVIRONMENT]->(e:Environment) "
            "RETURN di, collect(g) as genotypes, p, e"
        )

        for record in result:
            # Assuming we're interested in the first record only for this example
            di_node = record["di"]
            genotype_nodes = record["genotypes"]
            phenotype_node = record["p"]
            environment_node = record["e"]

            # Reconstruct the data instance
            data_instance["genotype"] = [
                genotype._properties for genotype in genotype_nodes
            ]
            data_instance["phenotype"] = phenotype_node._properties
            data_instance["environment"] = environment_node._properties

    driver.close()
    return data_instance


if __name__ == "__main__":
    uri = "neo4j://localhost:7687"
    username = "neo4j"
    password = "torchcell"
    add_data()
    # # dmf_dataset_0 = fetch_data_instance(uri, username, password)
    # # print(dmf_dataset_0)



# torchcell/adapters/kuzmin2018_adapter.py
# [[torchcell.adapters.kuzmin2018_adapter]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/adapters/kuzmin2018_adapter.py
# Test file: tests/torchcell/adapters/test_kuzmin2018_adapter.py

from tqdm import tqdm
import hashlib
import json
from biocypher._create import BioCypherEdge, BioCypherNode
from biocypher._logger import logger
from typing import Generator, Set
from torchcell.datasets.scerevisiae import (
    SmfKuzmin2018Dataset,
    DmfKuzmin2018Dataset,
    TmfKuzmin2018Dataset,
)
from torchcell.datamodels.schema import Genotype
from concurrent.futures import ProcessPoolExecutor, as_completed
import torch
from torchcell.adapters.cell_adapter import CellAdapter

# logger.debug(f"Loading module {__name__}.")
# logger.setLevel("INFO")


class SmfKuzmin2018Adapter(CellAdapter):
    def __init__(
        self,
        dataset: SmfKuzmin2018Dataset,
        compute_workers: int,
        io_workers: int,
        chunk_size: int = int(1e4),
        loader_batch_size: int = int(1e3),
    ):
        super().__init__(
            dataset, compute_workers, io_workers, chunk_size, loader_batch_size
        )
        self.dataset = dataset
        self.compute_workers = compute_workers
        self.io_workers = io_workers
        self.chunk_size = chunk_size
        self.loader_batch_size = loader_batch_size


class DmfKuzmin2018Adapter(CellAdapter):
    def __init__(
        self,
        dataset: DmfKuzmin2018Dataset,
        compute_workers: int,
        io_workers: int,
        chunk_size: int = int(1e4),
        loader_batch_size: int = int(1e3),
    ):
        super().__init__(
            dataset, compute_workers, io_workers, chunk_size, loader_batch_size
        )
        self.dataset = dataset
        self.compute_workers = compute_workers
        self.io_workers = io_workers
        self.chunk_size = chunk_size
        self.loader_batch_size = loader_batch_size


class TmfKuzmin2018Adapter(CellAdapter):
    def __init__(
        self,
        dataset: TmfKuzmin2018Dataset,
        compute_workers: int,
        io_workers: int,
        chunk_size: int = int(1e4),
        loader_batch_size: int = int(1e3),
    ):
        super().__init__(
            dataset, compute_workers, io_workers, chunk_size, loader_batch_size
        )
        self.dataset = dataset
        self.compute_workers = compute_workers
        self.io_workers = io_workers
        self.chunk_size = chunk_size
        self.loader_batch_size = loader_batch_size


if __name__ == "__main__":
    from biocypher import BioCypher
    from dotenv import load_dotenv
    from datetime import datetime
    import os
    import os.path as osp

    load_dotenv()
    time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    DATA_ROOT = os.getenv("DATA_ROOT")
    BIOCYPHER_CONFIG_PATH = os.getenv("BIOCYPHER_CONFIG_PATH")
    SCHEMA_CONFIG_PATH = os.getenv("SCHEMA_CONFIG_PATH")
    compute_workers = 8
    io_workers = 2

    ## Smf
    bc = BioCypher(
        output_directory=osp.join(DATA_ROOT, "database/biocypher-out", time),
        biocypher_config_path=BIOCYPHER_CONFIG_PATH,
        schema_config_path=SCHEMA_CONFIG_PATH,
    )
    dataset = SmfKuzmin2018Dataset(
        root=osp.join(DATA_ROOT, "data/torchcell/smf_kuzmin2018")
    )
    adapter = SmfKuzmin2018Adapter(
        dataset=dataset, compute_workers=compute_workers, io_workers=io_workers
    )
    bc.write_nodes(adapter.get_nodes())
    bc.write_edges(adapter.get_edges())
    bc.write_import_call()
    bc.write_schema_info(as_node=True)
    bc.summary()

    # ## Dmf
    bc = BioCypher(
        output_directory=osp.join(DATA_ROOT, "database/biocypher-out", time),
        biocypher_config_path=BIOCYPHER_CONFIG_PATH,
        schema_config_path=SCHEMA_CONFIG_PATH,
    )
    dataset = DmfKuzmin2018Dataset(osp.join(DATA_ROOT, "data/torchcell/dmf_kuzmin2018"))
    adapter = DmfKuzmin2018Adapter(
        dataset=dataset, compute_workers=compute_workers, io_workers=io_workers
    )
    bc.write_nodes(adapter.get_nodes())
    bc.write_edges(adapter.get_edges())
    bc.write_import_call()
    bc.write_schema_info(as_node=True)
    bc.summary()

    # ## Tmf
    bc = BioCypher(
        output_directory=osp.join(DATA_ROOT, "database/biocypher-out", time),
        biocypher_config_path=BIOCYPHER_CONFIG_PATH,
        schema_config_path=SCHEMA_CONFIG_PATH,
    )
    dataset = TmfKuzmin2018Dataset(osp.join(DATA_ROOT, "data/torchcell/tmf_kuzmin2018"))
    adapter = TmfKuzmin2018Adapter(
        dataset=dataset, compute_workers=compute_workers, io_workers=io_workers
    )
    bc.show_ontology_structure()
    bc.write_nodes(adapter.get_nodes())
    bc.write_edges(adapter.get_edges())
    bc.write_import_call()
    bc.write_schema_info(as_node=True)
    bc.summary()
    print()



# torchcell/prof.py
# [[torchcell.prof]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/prof.py
# Test file: torchcell/test_prof.py

import os
import os.path as osp
import time


def prof_input(func):
    import cProfile
    import datetime
    import pstats

    exp_name = input("Enter name of experiment: ")

    def inner(*args, **kwargs):
        dir = "profiles"
        if not osp.exists(dir):
            os.mkdir(dir)
        with cProfile.Profile() as pr:
            func(*args, **kwargs)
        stats = pstats.Stats(pr)
        now = datetime.datetime.now()
        now = now.strftime("%Y.%m.%d-%H.%M.%S")
        file_name = f"{exp_name}-{func.__name__}-{now}.prof"
        stats.dump_stats(filename=osp.join(dir, file_name))

    return inner


def prof(func):
    import cProfile
    import datetime
    import pstats

    def inner(*args, **kwargs):
        dir = "profiles"
        if not osp.exists(dir):
            os.mkdir(dir)
        with cProfile.Profile() as pr:
            func(*args, **kwargs)
        stats = pstats.Stats(pr)
        now = datetime.datetime.now()
        now = now.strftime("%Y.%m.%d-%H.%M.%S")
        file_name = f"experiment_name-{func.__name__}-{now}.prof"
        stats.dump_stats(filename=osp.join(dir, file_name))

    return inner


def main():
    @prof
    def test_func_dec():
        print("test func decorator")
        time.sleep(5)

    @prof_input
    def test_func_dec_in():
        print("test func decorator")
        time.sleep(5)

    test_func_dec()
    test_func_dec_in()

    # pip install snakeviz
    # In terminal run !snakeviz <path to .prof file>


if __name__ == "__main__":
    main()
